<?php
namespace SmartData;

if (!extension_loaded('cassandra'))
    die ("Unable to load cassandra extension\n");

require_once( __DIR__ . '/Config.php');
require_once( __DIR__ . '/Logger.php');
require_once( __DIR__ . '/Exception.php');
require_once( __DIR__ . '/Aggregators.php');
require_once( __DIR__ . '/Series.php');
require_once( __DIR__ . '/SmartData.php');
require_once( __DIR__ . '/Unit.php');
require_once( __DIR__ . '/Packer.php');
require_once( __DIR__ . '/Time.php');
require_once( __DIR__ . '/Anonymizer.php');

use SmartData\Anonymizer;
use SmartData\Utility\{Pack,Unpack};
use SmartData\Exception\{AuthenticationException,BadRequestException,InternalException};

class Backend_Common
{
    protected $_domain;
    protected $_username;
    protected $_password;

    protected $_client_ip;
    protected $_client_name;

    protected $_gw_id;

    private $_session_id;

    private $_internal;
    protected function internal(){ return $this->_internal; }

    protected const READ = 'r';
    protected const WRITE = 'w';

    // The parameter "$internal" means that the Backend has been called without a requisition.
    // Please, be careful and use only for development tasks.
    public function __construct(Credentials $credentials = NULL, $internal=false) {
        $credentials = $credentials ?? new Credentials();
        $testint = 0x00FF; // Ensure the machine is little-endian
        if(!($testint===current(unpack('v', pack('S', $testint))))) throw new \Exception("The machine is not little-endian");
        $this->_gw_id = -1; //NOTE: That means that the client isn't a known gateway
        $this->_internal = $internal; //NOTE: just for developers

        $USER   = ($credentials->username != NULL and $credentials->password != NULL);
        $DOMAIN = ($credentials->domain != NULL);

        // If it's not in an internal mode and the request is not secure (SSL), only the
        // domains listed in Config::config()::ACCESSIBLE_WITHOUT_SSL can be accessed.
        if (!$internal && !REQUEST_HTTPS && !in_array($credentials->domain, Config::config()::ACCESSIBLE_WITHOUT_SSL)){
            throw new AuthenticationException("Authentication failure [{$credentials}]");
        }

        switch (REQUEST_URI) {
            case 'search.php':
            case 'get.php':
                /**
                 *  ====== Summary of rules ======
                 *  HTTPS ^ CERT ^ domain -> OK, domain = $domain
                 *  HTTPS ^ CERT  -> OK, domain = DB(CERT)->domain
                 *  HTTPS ^ domain ^ valid(usr+password) -> OK, domain = $domain
                 *  HTTPS ^ valid(usr+password) -> OK, domain = public
                 *  HTTPS ^ domain -> OK, domain is public ? domain = $domain : fail
                 *  HTTPS -> OK, domain = public
                 *  HTTP ^ valid(usr+password) -> OK, domain = hydrology
                 *  Fail
                 */

                if ( REQUEST_HTTPS && REQUEST_CERT && $DOMAIN ) {
                    if(!$this->_checkCertificate(REQUEST_SERIAL, self::READ, $credentials->domain))
                        throw new AuthenticationException(REQUEST_DBG.' crd:'.$credentials.'(CD) Unauthorized access');
                } else if ( REQUEST_HTTPS && REQUEST_CERT ) {
                    if(!$this->_checkCertificate(REQUEST_SERIAL, self::READ))
                        throw new AuthenticationException(REQUEST_DBG.' crd:'.$credentials.'(C) Unauthorized access');
                } else if ( REQUEST_HTTPS && $DOMAIN && $USER ) {
                    if(!$this->_checkUser($credentials->username, $credentials->password, self::READ, $credentials->domain))
                        throw new AuthenticationException(REQUEST_DBG.' crd:'.$credentials.'(DU) Unauthorized access');
                } else if ( REQUEST_HTTPS && $USER ) {
                    $credentials->domain = Config::config()::PUBLIC_DOMAIN;
                    if(!$this->_checkUser($credentials->username, $credentials->password, self::READ, Config::config()::PUBLIC_DOMAIN))
                        throw new AuthenticationException(REQUEST_DBG.' crd:'.$credentials.'(U) Unauthorized access');
                } else if ( REQUEST_HTTPS && $DOMAIN ) {
                    // NOTE: Publicly readable domains
                    if(!$this->_checkUser(Config::config()::DEFAULT_USERNAME, Config::config()::DEFAULT_PASSWORD, self::READ, $credentials->domain))
                        throw new AuthenticationException(REQUEST_DBG.' crd:'.$credentials.'(D) Unauthorized access');
                } else if ( REQUEST_HTTPS ) {
                    $this->_domain   = Config::config()::PUBLIC_DOMAIN;
                    $this->_username = Config::config()::DEFAULT_USERNAME;
                    $this->_password = Config::config()::DEFAULT_PASSWORD;
                } else if ( !REQUEST_HTTPS && $DOMAIN && $USER ) {
                    //NOTE: Domain that are accessible with HTTP (without SSL)
                    if(!$this->_checkUser($credentials->username, $credentials->password, self::READ, $credentials->domain))
                        throw new AuthenticationException(REQUEST_DBG.' crd:'.$credentials.'(DU) Unauthorized access');
                } else {
                    throw new BadRequestException(REQUEST_DBG.' crd:'.$credentials.' Bad protocol and options');
                }
                break;
            case 'create.php'    :
            case 'attach.php'    :
            case 'put.php'       :
            case 'batch.php'     :
            case 'put_batch.php' :
            case 'mv_put.php'    :
            case 'mu_put.php'    :
            case 'md_put.php'    :
            case 'finish.php'    :
            case 'describe.php'  :
            case 'list.php'      :
            case 'context.php'   :
                /**
                 *  ====== Summary of rules ======
                 *  HTTPS ^ CERT ^ domain -> OK, domain = $domain
                 *  HTTPS ^ CERT  -> OK, domain = DB(CERT)->domain
                 *  HTTPS ^ domain ^ valid(usr+password) -> OK, domain = $domain
                 *  HTTPS ^ valid(usr+password) -> OK, domain = public
                 *  HTTP  ^ valid(usr+password) -> OK, domain = hydrology
                 *  Fail
                 */

                if ( REQUEST_HTTPS && REQUEST_CERT && $DOMAIN ) {
                    if(!$this->_checkCertificate(REQUEST_SERIAL, self::WRITE, $credentials->domain))
                        throw new AuthenticationException(REQUEST_DBG.' crd:'.$credentials.'(CD) Unauthorized access');
                } else if ( REQUEST_HTTPS && REQUEST_CERT ) {
                    if(!$this->_checkCertificate(REQUEST_SERIAL, self::WRITE))
                        throw new AuthenticationException(REQUEST_DBG.' crd:'.$credentials.'(C) Unauthorized access');
                } else if ( REQUEST_HTTPS && $DOMAIN && $USER ) {
                    if(!$this->_checkUser($credentials->username, $credentials->password, self::WRITE, $credentials->domain))
                        throw new AuthenticationException(REQUEST_DBG.' crd:'.$credentials.'(DU) Unauthorized access');
                } else if ( REQUEST_HTTPS && $USER ) {
                    //TODO: Check what is the main idea
                    $credentials->domain = Config::config()::PUBLIC_DOMAIN;
                    if(!$this->_checkUser($credentials->username, $credentials->password, self::WRITE, Config::config()::PUBLIC_DOMAIN))
                        throw new AuthenticationException(REQUEST_DBG.' crd:'.$credentials.'(U) Unauthorized access');
                } else if ( REQUEST_HTTPS && $DOMAIN ) {
                    // NOTE: Public writable domains
                    if(!$this->_checkUser(Config::config()::DEFAULT_USERNAME, Config::config()::DEFAULT_PASSWORD, self::WRITE, $credentials->domain))
                        throw new AuthenticationException(REQUEST_DBG.' crd:'.$credentials.'(D) Unauthorized access');
                } else if ( !REQUEST_HTTPS && $DOMAIN && $USER ) {
                    //NOTE: Domain that are accessible with HTTP (without SSL)
                    if(!$this->_checkUser($credentials->username, $credentials->password, self::WRITE, $credentials->domain))
                       throw new AuthenticationException(REQUEST_DBG.' crd:'.$credentials.'(DU) Unauthorized access');
                } else {
                    throw new BadRequestException(REQUEST_DBG.' crd:'.$credentials.' Bad protocol and options');
                }
                break;
            default:
                if($this->internal()){
                    $this->_domain = $credentials->domain;
                    $this->_username = $credentials->username;
                    $this->_password = $credentials->password;
                } else
                    throw new BadRequestException(REQUEST_DBG.' crd:'.$credentials.' Bad protocol and options');
                break;
        }

        if($this->_domain == null || $this->_username == null || $this->_password == null){
            throw new BadRequestException("Incorrect use of protocol and options {$credentials}");
        }

        $this->_session_id = $this->_domain.'_'.substr(uniqid(), -5);
        if(REQUEST_URI != 'get.php')
            self::debug_X($this);
    }

    // Checks the health of databases
    public static function health_check() : bool {
        $sucess = true;
        try {
            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME,
                                        Config::config()::MYSQL_PORT,
                                        Config::config()::MYSQL_DBNAME,
                                        Config::config()::MYSQL_USERNAME,
                                        Config::config()::MYSQL_PASSWORD);
            $conn = null;
            $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME,
                                               Config::config()::CASSANDRA_PORT,
                                               Config::config()::DEFAULT_USERNAME,
                                               Config::config()::DEFAULT_PASSWORD,
                                               Config::config()::CASSANDRA_KEYSPACE);
            $session->close();
        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }
        return $sucess;
    }

    // Check Certificate
    private function _checkCertificate(string $certificate, string $level, string $domain = null) : bool {
        $sucess = true;
        $target_domain = "";
        $gw_id = -1;

        $parameters = array(':certificate' => $certificate,
                            ':level'       => $level);


        try {

            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, Config::config()::MYSQL_PASSWORD);
            if(Config::config()::api_version() == Config::config()::VERSION_1_1) // TODO: Remove from development_mode
                $table = 'clients';
            else
                $table = 'gateways'; // table that has to be changed to clients

            if($domain == null){
                $stmt = $conn->prepare("SELECT id, domain FROM {$table} WHERE enable = true AND certificate = :certificate AND INSTR(level, :level) > 0 ORDER BY id ASC LIMIT 1");
            } else {
                $parameters = array_merge($parameters, array(':domain'=> $domain));
                $stmt = $conn->prepare("SELECT id, domain FROM {$table} WHERE enable = true AND certificate = :certificate AND (domain = :domain OR domain = '*') AND INSTR(level, :level) > 0");
            }

            $stmt->execute($parameters);
            if($stmt->rowCount() != 1)
                throw new AuthenticationException("Permission not found or number of domains larger than one [{$certificate} / {$level}]");
            else{
                $result = $stmt->fetch(\PDO::FETCH_OBJ);
                $target_domain = $domain ?? $result->domain;
                $gw_id = $result->id;
            }

        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            $conn = null;
        }

        if ($sucess) {
            $this->_domain = $target_domain;
            $this->_username = Config::config()::CASSANDRA_SUPERUSER;
            $this->_password = Config::config()::CASSANDRA_SUPERPASS;
            $this->_gw_id = $gw_id;
            return true;
        } else {
            return false;
        }
    }

    // Check User Connection
    private function _checkUserConnection(string $username, string $password) : bool {
        $sucess = false;
        try {
            $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, $username, $password, Config::config()::CASSANDRA_KEYSPACE);
            $sucess = true;
        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
        }
        return $sucess;
    }

    // Check if an user has permission to read/write on a domain
    private function _checkUser(string $username, string $password, string $level, string $domain = null) : bool {
        //TODO: Check what is the main idea
        if($domain == NULL) return false;

        $sucess = false;

        try {
            $permission = array();
            switch($level) {
                case self::READ:
                    array_push($permission, 'SELECT');
                case self::WRITE:
                    array_push($permission, 'MODIFY');
                    break; // Test to write
                default; return false; // invalid option
            }

            if(Config::config()::api_version() == Config::config()::VERSION_1_1) { // TODO: Remove from development_mode
                $keyspace = Config::config()::CASSANDRA_KEYSPACE;
                //$query = "LIST ALL PERMISSIONS ON table {$keyspace}.{$domain} OF {$username};";
                $query = "LIST ALL PERMISSIONS OF {$username};";
                $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, Config::config()::CASSANDRA_SUPERUSER, Config::config()::CASSANDRA_SUPERPASS, Config::config()::CASSANDRA_KEYSPACE);
                $statement = new \Cassandra\SimpleStatement($query);
                $result = $session->execute($statement);

                foreach ($result as $column) {
                    $pattern = "<table {$keyspace}.{$domain}_v";
                    if((strpos($column['resource'], $pattern) !== false)) {
                        if(in_array($column['permission'], $permission)){
                            $sucess = true;
                            break;
                        }
                    }
                }
                if($sucess)
                    $sucess = $this->_checkUserConnection($username, $password);
            } else {
                $keyspace = Config::config()::CASSANDRA_KEYSPACE;
                $query = "LIST ALL PERMISSIONS ON table {$keyspace}.{$domain} OF {$username};";

                $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, Config::config()::CASSANDRA_SUPERUSER, Config::config()::CASSANDRA_SUPERPASS, Config::config()::CASSANDRA_KEYSPACE);
                $statement = new \Cassandra\SimpleStatement($query);
                $result = $session->execute($statement);

                foreach ($result as $column) {
                    if(in_array($column['permission'], $permission)){
                        $sucess = $this->_checkUserConnection($username, $password);
                        break;
                    }
                }
            }

        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
        }

        if ($sucess) {
            $this->_domain = $domain;
            $this->_username = $username;
            $this->_password = $password;
            return true;
        } else {
            return false;
        }
    }

    // Open a connection with MySQL database
    protected static function _mysqlConnect(string $servername, int $port, string $dbname, string $username, string $password) : \PDO {
        $conn = new \PDO("mysql:host=$servername;port=$port;dbname=$dbname", $username, $password);
        // set the PDO error mode to exception
        $conn->setAttribute(\PDO::ATTR_ERRMODE, \PDO::ERRMODE_EXCEPTION);

        return $conn;
    }

    // Open a connection with Cassandra database
    protected static function _cassandraConnect(string $servername, int $port, string $username, string $password, string $keyspace) : \Cassandra\DefaultSession {
        $cluster = \Cassandra::cluster()->withContactPoints($servername)
                                        ->withPort($port)
                                        ->withCredentials($username, $password)
                                        //->withDefaultPageSize(200)
                                        ->build();
        $session = $cluster->connect($keyspace);
        return $session;
    }

    protected function debug_X(string $message, $ctrl=false) {
        Logger::debug_X('['.$this->_session_id.'] '.$message, ($ctrl === false) ? get_called_class() : $ctrl);
    }

    public function __toString() { return REQUEST_DBG.' crd:{'.$this->_username.'@'.$this->_domain.'}'; }

    public function domainInUse():string {
        return $this->_domain;
    }
}

class Tracker
{
    public $version;
    public $unit;
    public $signature;
    public $dev;
    public $t0;
    public $t1;
    public $count;
    public $workflow;
    public $domain;
    public $row_id;
    public $type;
    public $period;
    public $event; 
    public $uncertainty; 
    public $id;
    
    public function __construct($version, $unit, $signature, $t0, $t1, $domain = null, $dev = 0, 
           $workflow = 0, $type = 'OLD', $period = 0, $uncertainty = 0, $id = 0, $row_id = 0, $event = null) {
        $ini = $t0 ?? 0;
        $end = $t1 ?? 999999999999999;
        if ($ini == 0 or $t0 > $t1 or (false)){
            throw new BadRequestException("Invalid tracker");
        }

        $this->version   = $version;
        $this->unit      = $unit;
        $this->signature = $signature;
        $this->t0        = $ini;
        $this->t1        = $end;
        $this->domain    = $domain;
        $this->dev       = $dev;
        $this->workflow  = $workflow;
        $this->type      = $type;
        $this->period    = $period;
        $this->uncertainty = $uncertainty;
        $this->row_id    = $row_id;
        $this->event     = $event;   
        $this->id        = $id;
    }
}

final class Credentials
{
    public $domain;
    public $username;
    public $password;

    public function __construct($d=null, $u=null, $p=null) {
        if ($u == null xor $p == null) {
            throw new Exception\BadCredentialsException("Invalid credentials fields [{$username}@{$domain}]");
        }
        $this->domain   = $d;
        $this->username = $u;
        $this->password = $p;
    }

    public function pack(array $suppress = array()) {
        $bin  = '';
        $bin .= Pack::string($this->domain);
        $bin .= Pack::string($this->username);
        $bin .= Pack::string($this->password);
        return $bin;
    }

    public static function unpack(& $bin, array $compl = array()) {
        $domain   = Unpack::string($bin, true);
        $username = Unpack::string($bin, true);
        $password = Unpack::string($bin, true);
        return new self($domain, $username, $password);
    }

    public static function fromJson(\stdClass $json) {
        return new self($json->domain   ?? null, $json->username ?? null, $json->password ?? null);
    }

    public function __toString() {
        $string  = "{";
        $string .= "{$this->username}@{$this->domain}, ";
        $string .= strlen($this->password);
        $string .= '}';
        return $string;
    }
}

## Backend of API version 1.0 ##############################################################################
class Backend_V1_0 extends Backend_Common
{
    // Create a new series if not exist
    public function create($series) {
        $sucess = true;
        $parameters = array(':version'  => $series->version,
                            ':unit'     => $series->unit,
                            ':x'        => $series->x,
                            ':y'        => $series->y,
                            ':z'        => $series->z,
                            ':r'        => $series->r,
                            ':t0'       => $series->t0,
                            ':t1'       => $series->t1,
                            ':domain'   => $this->_domain,
                            ':count'    => Config::config()::CASSANDRA_MAX_ROW_SIZE);
        try {    
            Unit::validate( $series->unit );
            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, Config::config()::MYSQL_PASSWORD);
            $table = Config::config()::MySQL_Table($series->version);
            $stmt_select = $conn->prepare("SELECT * FROM {$table} WHERE version = :version
                                              AND unit = :unit
                                              AND x = :x
                                              AND y = :y
                                              AND z = :z
                                              AND r = :r
                                              AND t0 = :t0
                                              AND t1 = :t1
                                              AND domain = :domain
                                              AND count < :count
                                            ORDER BY id DESC LIMIT 1;");
            $stmt_select->execute($parameters);
            $series = $stmt_select->fetch(\PDO::FETCH_OBJ);
            if ($series == NULL) {
                // Not exists yet
                $stmt_insert = $conn->prepare("INSERT INTO {$table} VALUES (DEFAULT, :version, :unit, :x, :y, :z, :r, :t0, :t1, DEFAULT, :domain, (SELECT counter(:domain)), DEFAULT);");
                $stmt_insert->execute($parameters);
            }
        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            $conn = null;
        }

        return $sucess;
    }
    // Create a new series if there is not one that contains it
    public function attach($series) {
        $sucess = true;

        $table = Config::config()::MySQL_Table($series->version);

        if($series->version == SmartData::MOBILE_VERSION){
            $parameters = array(':version'   => $series->version,
                                ':unit'      => $series->unit,
                                ':signature' => $series->signature,
                                ':sigsha'    => hash('sha256',$series->signature),
                                ':x'         => $series->x,
                                ':y'         => $series->y,
                                ':z'         => $series->z,
                                ':r'         => $series->r,
                                ':t0'        => $series->t0,
                                ':t1'        => $series->t1,
                                ':domain'    => $this->_domain,
                                ':count'     => Config::config()::CASSANDRA_MAX_ROW_SIZE);

            $query = "SELECT * FROM {$table} WHERE version = :version
                  AND unit = :unit
                  AND signature in ( :signature , :sigsha )
                  AND r >= (SQRT(POW(x - :x, 2) + POW(y - :y, 2) + POW(z - :z, 2)) + :r)
                  AND t0 <= :t0 AND t1 >= :t1
                  AND domain = :domain
                  AND count < :count
                ORDER BY r ASC LIMIT 1;";

            $insert = "INSERT INTO {$table} VALUES (DEFAULT, :version, :unit, :x, :y, :z, :r, :t0, :t1, DEFAULT, :domain, (SELECT counter(:domain)), DEFAULT, :sigsha);";

            Logger::debug('Attaching MOBILE: '.$series, true);

        } else {
            Logger::debug('Attaching STATIC: '.$series, true);
            $parameters = array(':version'   => $series->version,
                                ':unit'      => $series->unit,
                                ':x'         => $series->x,
                                ':y'         => $series->y,
                                ':z'         => $series->z,
                                ':r'         => $series->r,
                                ':t0'        => $series->t0,
                                ':t1'        => $series->t1,
                                ':domain'    => $this->_domain,
                                ':count'     => Config::config()::CASSANDRA_MAX_ROW_SIZE);

            $query = "SELECT * FROM {$table} WHERE version = :version
                  AND unit = :unit
                  AND r >= (SQRT(POW(x - :x, 2) + POW(y - :y, 2) + POW(z - :z, 2)) + :r)
                  AND t0 <= :t0 AND t1 >= :t1
                  AND domain = :domain
                  AND count < :count
                ORDER BY r ASC LIMIT 1;";

            $insert = "INSERT INTO {$table} VALUES (DEFAULT, :version, :unit, :x, :y, :z, :r, :t0, :t1, DEFAULT, :domain, (SELECT counter(:domain)), DEFAULT, DEFAULT);";
        }

        try {

            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME,
                                        Config::config()::MYSQL_PORT,
                                        Config::config()::MYSQL_DBNAME,
                                        Config::config()::MYSQL_USERNAME,
                                        Config::config()::MYSQL_PASSWORD);

            $stmt_select = $conn->prepare($query);
            $stmt_select->execute($parameters);
            $series = $stmt_select->fetch(\PDO::FETCH_OBJ);

            if ($series == NULL) {
                $stmt_insert = $conn->prepare($insert);
                $stmt_insert->execute($parameters);
                Logger::debug(__FILE__."> Attach (new series)");
            }
        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            $conn = null;
        }

        return $sucess;
    }
    // Insert SmartData into series/tracker matched
    public function insert(SmartData $smart_data) {
        if ($smart_data instanceof StaticSmartData)
            return $this->_insertStaticSmartData($smart_data);
        else if ($smart_data instanceof MobileSmartData)
            return $this->_insertMobileSmartData($smart_data);
        else
            throw new BadRequestException("Error processing JSON request: unsupported SmartData version");
    }
    // Insert a static version of smartdata
    private function _insertStaticSmartData(StaticSmartData $smart_data) : bool {
        $sucess = true;

        $parameters = array(':unit'       => $smart_data->unit,
                            ':x'          => $smart_data->x,
                            ':y'          => $smart_data->y,
                            ':z'          => $smart_data->z,
                            ':time'       => $smart_data->time,
                            ':domain'     => $this->_domain);

        $is_digital = $smart_data->unit()->is_digital();

        try {

            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, Config::config()::MYSQL_PASSWORD);
            $table = Config::config()::MySQL_Table($smart_data->version);
            $series = NULL;
            $first = true;
            do {
                $stmt_select = $conn->prepare("SELECT * FROM {$table} WHERE unit = :unit
                                                              AND (SQRT(POW(x - :x, 2) + POW(y - :y, 2) + POW(z - :z, 2))) <= r
                                                              AND :time >= t0 AND :time <= t1
                                                              AND domain = :domain
                                                            ORDER BY r ASC, count ASC LIMIT 1;");

                $stmt_select->execute($parameters);
                $series = $stmt_select->fetch(\PDO::FETCH_OBJ);

                if ($first && $series!=NULL && $series->count>=Config::config()::CASSANDRA_MAX_ROW_SIZE) {
                    $this->create(new Series($series->version,
                                      $series->unit,
                                      $series->x,
                                      $series->y,
                                      $series->z,
                                      $series->r,
                                      $series->t0,
                                      $series->t1));
                    $first = false;
                } else {
                    break;
                }
            } while (true);

            if ($series == NULL) {
                throw new Exception\NoMatchingSeriesException('No matching series');
            }else{
                $entry_type = \Cassandra\Type::userType(
                    // Just for attend projects requirements of SO2 -------------------------------------
                    'value',      ($is_digital) ? \Cassandra\Type::blob() : \Cassandra\Type::double(),
                    // ----------------------------------------------------------------------------------
                    'error',      \Cassandra\Type::tinyint(),
                    'confidence', \Cassandra\Type::tinyint(),
                    'x',          \Cassandra\Type::int(),
                    'y',          \Cassandra\Type::int(),
                    'z',          \Cassandra\Type::int(),
                    'd',          \Cassandra\Type::bigint(),
                    'gw',         \Cassandra\Type::bigint()
                );

                $smart_data->confidence = $this->assign_confidence($smart_data);

                $arguments = array(
                    new \Cassandra\Bigint($series->slice),
                    new \Cassandra\Varint($smart_data->time),
                    $entry_type->create(
                        'value',      ($is_digital) ? new \Cassandra\Blob($smart_data->value) : (double) $smart_data->value,
                        'error',      new \Cassandra\Tinyint($smart_data->error),
                        'confidence', new \Cassandra\Tinyint($smart_data->confidence),
                        'x',          (int)$smart_data->x,
                        'y',          (int)$smart_data->y,
                        'z',          (int)$smart_data->z,
                        'd',          new \Cassandra\Bigint($smart_data->dev??0), //TODO: remove the ??
                        //'gw',         new \Cassandra\Bigint($smart_data->gw->id)
                        'gw',         new \Cassandra\Bigint($this->_gw_id) // TODO: CHECK THIS
                    )
                );

                $cassandra_table = "{$this->_domain}" . (($is_digital) ? '_dig' : '');

                $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, $this->_username, $this->_password, Config::config()::CASSANDRA_KEYSPACE);
                $insert_stmt = $session->prepare("INSERT INTO {$cassandra_table} (id, timestamp, entry) VALUES (?, ?, ?);");
                $session->execute($insert_stmt, array('arguments' => $arguments));

                $update_stmt = $conn->prepare("UPDATE {$table} SET count = count + 1 WHERE id = :id LIMIT 1;");
                $update_stmt->execute(array('id' => $series->id));
           }

        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }
        return $sucess;
    }
    // Insert a mobile version of smartdata
    private function _insertMobileSmartData(MobileSmartData $smart_data) : bool {

        $sucess = true;

        $parameters = array(':unit'       => $smart_data->unit,
                            ':signature'  => $smart_data->signature,
                            ':sigsha'     => hash('sha256',$smart_data->signature),
                            ':x'          => $smart_data->x,
                            ':y'          => $smart_data->y,
                            ':z'          => $smart_data->z,
                            ':time'       => $smart_data->time,
                            ':domain'     => $this->_domain);

        $is_digital = $smart_data->unit()->is_digital();

        try {

            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, Config::config()::MYSQL_PASSWORD);
            $table = Config::config()::MySQL_Table($smart_data->version);
            $series = NULL;
            $first = true;
            do {
                $stmt_select = $conn->prepare(
                    "SELECT * FROM {$table} WHERE unit = :unit 
                    AND (SQRT(POW(x - :x, 2) + POW(y - :y, 2) + POW(z - :z, 2))) <= r 
                    AND :time >= t0 AND :time <= t1 
                    AND domain = :domain 
                    AND signature in ( :signature , :sigsha ) 
                    ORDER BY r ASC, count ASC LIMIT 1;");

                $stmt_select->execute($parameters);
                $series = $stmt_select->fetch(\PDO::FETCH_OBJ);

                if ($first && $series!=NULL && $series->count>=Config::config()::CASSANDRA_MAX_ROW_SIZE) {
                    $this->create(
                        new Series($series->version,
                            $series->unit,
                            $series->x,
                            $series->y,
                            $series->z,
                            $series->r,
                            $series->t0,
                            $series->t1,
                            $series->dev //TODO verificar pq isto funcionava sem dev
                        ));
                    $first = false;
                } else {
                    break;
                }
            } while (true);

            if ($series == NULL) {
                throw new Exception\NoMatchingSeriesException('No matching series');
            }else{

                $entry_type = \Cassandra\Type::userType(
                    // Just for attend projects requirements of SO2 -------------------------------------
                    'value',      ($is_digital) ? \Cassandra\Type::blob() : \Cassandra\Type::double(),
                    // ----------------------------------------------------------------------------------
                    'error',      \Cassandra\Type::tinyint(),
                    'confidence', \Cassandra\Type::tinyint(),
                    'x',          \Cassandra\Type::int(),
                    'y',          \Cassandra\Type::int(),
                    'z',          \Cassandra\Type::int(),
                    'd',          \Cassandra\Type::bigint(),
                    'gw',         \Cassandra\Type::bigint()
                );

                $smart_data->confidence = $this->assign_confidence($smart_data);

                $arguments = array(
                    new \Cassandra\Bigint($series->slice),
                    new \Cassandra\Varint($smart_data->time),
                    $entry_type->create(
                        'value',      ($is_digital) ? new \Cassandra\Blob($smart_data->value) : (double) $smart_data->value,
                        'error',      new \Cassandra\Tinyint($smart_data->error),
                        'confidence', new \Cassandra\Tinyint($smart_data->confidence),
                        'x',          (int)$smart_data->x,
                        'y',          (int)$smart_data->y,
                        'z',          (int)$smart_data->z,
                        'd',          new \Cassandra\Bigint($smart_data->dev??0), //TODO: remove the ??
                        //'gw',         new \Cassandra\Bigint($smart_data->gw->id)
                        'gw',         new \Cassandra\Bigint($this->_gw_id) // TODO: CHECK THIS
                    )
                );

                $cassandra_table = "{$this->_domain}" . (($is_digital) ? '_dig' : '');

                $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, $this->_username, $this->_password, Config::config()::CASSANDRA_KEYSPACE);
                $insert_stmt = $session->prepare("INSERT INTO {$cassandra_table} (id, timestamp, entry) VALUES (?, ?, ?);");
                $session->execute($insert_stmt, array('arguments' => $arguments));

                $update_stmt = $conn->prepare("UPDATE {$table} SET count = count + 1 WHERE id = :id LIMIT 1;");
                $update_stmt->execute(array('id' => $series->id));
            }

        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }

        return $sucess;
    }
    // Query data from series
    public function query(Series $series, $aggr=null, $options=null) {
        $response_json = NULL;

        try {

            $parameters = array(':version'  => $series->version,
                                ':t0'       => $series->t0,
                                ':t1'       => $series->t1,
                                ':domain'   => $this->_domain);

            $table = Config::config()::MySQL_Table($series->version);
            $query = "SELECT slice, unit FROM {$table}
                      WHERE version = :version AND ( (:t0 >= t0 AND :t0 <= t1) OR (:t0 <= t0 AND :t1 >= t1) OR (:t1 >= t0 AND :t1 <= t1 ) )";

            if($series->unit != null){
                $parameters = array_merge($parameters, array(':unit' => $series->unit));
                //$parameters = array_merge($parameters, array(':dev' => $series->dev));
                $query .= ' AND unit = :unit';
                //$query .= ' AND dev = :dev';
            }

            $invalid_sphere = (int)($series->x === null) +
                              (int)($series->y === null) +
                              (int)($series->z === null) +
                              (int)($series->r === null);

            if ($invalid_sphere != 0 && $invalid_sphere != 4) {
                throw new BadRequestException("Invalid sphere parameters");
            } else if($invalid_sphere == 0) {
                $parameters = array_merge($parameters, array(':x'    => $series->x,
                                                             ':y'    => $series->y,
                                                             ':z'    => $series->z,
                                                             ':r'    => $series->r));
                $query .= ' AND SQRT(POW(x - :x, 2) + POW(y - :y, 2) + POW(z - :z, 2)) <= :r + r';
                $filter_sphere = true;
            } else {
                $filter_sphere = false;
            }

            $query .= ' AND domain = :domain AND count > 0;';

            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME,
                                        Config::config()::MYSQL_PORT,
                                        Config::config()::MYSQL_DBNAME,
                                        Config::config()::MYSQL_USERNAME,
                                        Config::config()::MYSQL_PASSWORD);

            $select_stmt = $conn->prepare($query);
            $select_stmt->execute($parameters);

            $id_set = \Cassandra\Type::collection(\Cassandra\Type::bigint())->create();

            $selected_slices = array_column($select_stmt->fetchAll(), 'unit', 'slice');
            foreach ($selected_slices as $slice => $unit) {
                $id_set->add(new \Cassandra\Bigint($slice));
            }

            $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, $this->_username, $this->_password, Config::config()::CASSANDRA_KEYSPACE);

            $is_digital  = $series->unit()->is_digital();
            if($is_digital) {
                $type = $_GET['type'] ?? $options->type ?? null;
                $type = $type << 29;
            }

            $cassandra_table = "{$this->_domain}" . (($is_digital) ? '_dig' : '');
            $points_limit = Config::config()::POINTS_LIMIT;
            $cass_select_stmt = $session->prepare("SELECT * FROM {$cassandra_table} WHERE (id IN ?) AND (timestamp >= ?) AND (timestamp <= ?) limit {$points_limit};");

            $result_page = $session->execute($cass_select_stmt,
                            array('arguments' => array($id_set, new \Cassandra\Varint($series->t0), new \Cassandra\Varint($series->t1))));

            //ini_set('memory_limit', '-1'); // check this
            //ini_set('max_execution_time', 300); // check this

            $index = 0;
            $response_json = array('series' => array());

            $aggregator = ($this->internal()) ? null : Aggregator::new($aggr);

            while($result_page) {
                foreach ($result_page as $column) {
                    $slice = (int)$column['id'];
                    $entry = $column['entry'];
                    $time  = (string)$column['timestamp']->value(); //bigint
                    if($is_digital){
                        $value_blob = $entry->get('value')->toBinaryString();
                        switch ($type) {
                            case Unit::I32:
                                $value = Unpack::uInt32($value_blob);
                                break;
                            case Unit::I64:
                                $value = Unpack::uInt64($value_blob);
                                break;
                            case Unit::F32:
                                $value = Unpack::f32($value_blob);
                                break;
                            case Unit::D64:
                                break;
                            default:
                                $value = Unpack::d64($value_blob);
                                break;
                        }

                        if($value === null){
                            if($value_blob != null)
                                $value = $value_blob; // TODO: remove this in the new API
                            else
                                $value = -400; // TODO: remover isso quando valores inválidos não puderem mais ser inseridos
                        }
                    } else {
                        $value = (double)$entry->get('value'); // double
                    }

                    $error = (int)$entry->get('error');             //tinyint
                    $conf  = (int)$entry->get('confidence');        //tinyint
                    $x     = (int)$entry->get('x');                 //int
                    $y     = (int)$entry->get('y');                 //int
                    $z     = (int)$entry->get('z');                 //int
                    $d     = (int)$entry->get('d');                 //int
                    $gw    = (int)$entry->get('gw');                //int (TO DO GET GATEWAY)

                    if ( $filter_sphere && !$series->contains_point($x, $y, $z) )
                        continue;

                    if ($series->unit == 2226272548 && $d != $series->dev) //Angle Rate Sensor (testing)
                        continue;

                    if ($this->_domain == 'tutorial' && $series->unit == 2224179556 && $d != $series->dev) //Debuging
                        continue;

                    $smartdata = null;

                    switch ($series->version) {
                        case SmartData::STATIC_VERSION:
                            $smartdata = new StaticSmartData($series->unit,$value,$error,$conf,$x,$y,$z,$time,$d,$gw);
                            break;
                        case SmartData::MOBILE_VERSION:
                            $smartdata = new MobileSmartData($series->unit,$value,$error,$conf,$x,$y,$z,$time,$d,$gw, null);
                            break;
                        default:
                            continue 2;
                            break;
                    }

                    if($aggregator != null) {
                        $smartdata = $aggregator->aggregate($smartdata);
                    }

                    if($smartdata != null) {
                        if(is_array($smartdata)){
                            foreach ($smartdata as $key => $data) {
                                $response_json['series'][$index++] = $data->toArray();
                                if ($index >= Config::config()::POINTS_LIMIT) break;
                            }
                        } else {
                            $response_json['series'][$index++] = $smartdata->toArray();
                            if($this->internal()){
                                $response_json['series'][$index-1]['slice'] = $slice;
                                $response_json['series'][$index-1]['gw'] = $gw;
                            }
                        }
                    } else {
                        continue;
                    }

                    if ($index >= Config::config()::POINTS_LIMIT) break 2;
                }

                $result_page = $result_page->nextPage();

                if(!$result_page && isset($aggregator) && $index < Config::config()::POINTS_LIMIT){
                    $smartdata = $aggregator->finish();
                    if($smartdata != null) {
                        $response_json['series'][$index++] = $smartdata->toArray();
                    }
                }
            }

            $session->close();

            usort($response_json['series'], function ($a, $b) {
                $a_timestamp  = gmp_init($a['timestamp']);
                $b_timestamp  = gmp_init($b['timestamp']);
                return gmp_cmp($a_timestamp, $b_timestamp);
            });

        } catch (\Exception $e) {
            Logger::exception($e);
            $response_json = NULL;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }

        return $response_json;
    }

    // This method inserts multiple datapoints from the same location and unit into the database. 
    // May be of the same device, or from multiple devices. 
    public function insert_multiple(SmartData ...$smart_data) {
        $sucess = true;

        $smart_data_count = count($smart_data);

        if (count < 1) {
            error_log('List of smartdata is empty....');
            return;
        }

        // First: define max and min timestamp, and verify if a there is a series that contains all

        $sd =  $smart_data[0];
        $min = $sd[0]->time;
        $max = $min;
        foreach ($smart_data as $sd) {
            if ($sd->time > $max) {
                $max = $sd->time;
            } else if ($sd->time < $min) {
                $min = $sd->time; 
            }
        }

        try {

            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME,
                                        Config::config()::MYSQL_PORT,
                                        Config::config()::MYSQL_DBNAME,
                                        Config::config()::MYSQL_USERNAME,
                                        Config::config()::MYSQL_PASSWORD);

            $table = Config::config()::MySQL_Table($series->version);

            $series_db = NULL;

            $parameters = array(':version'   => $sd->version,
                                ':unit'      => $sd->unit,
                                ':x'         => $sd->x,
                                ':y'         => $sd->y,
                                ':z'         => $sd->z,
                                ':t0'        => $min,
                                ':t1'        => $max,
                                ':domain'    => $this->_domain,
                                ':count'     => Config::config()::CASSANDRA_MAX_ROW_SIZE );

           $query = "SELECT * FROM {$table} WHERE version = :version
                        AND unit = :unit 
                        AND signature in ( :signature , :sigsha ) 
                        AND r >= (SQRT(POW(x - :x, 2) + POW(y - :y, 2) + POW(z - :z, 2))) 
                        AND t0 <= :t0 AND t1 >= :t1 
                        AND domain = :domain 
                        AND count < :count 
                        ORDER BY r ASC LIMIT 1;";

            $stmt_select = $conn->prepare( $query );
            $stmt_select->execute($parameters);
            $sdb = $stmt_select->fetch(\PDO::FETCH_OBJ);
            $series = new Series( $sdb->version, $sdb->unit, $sdb->x, $sdb->y, $sdb->z, $sdb->r, $sdb->t0, $sdb->t1, $sdb->dev);
            $sucess = $this->_insertStaticSmartDataBatch($series, ...$smart_data);
            error_log('Inserted '.$smart_data_count.' MSD records in DB');
        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
            error_log('Failed to insert MSD'.$e);
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }

        return $sucess;
    }

    // Track mobile node
    public function track(Series $series, $aggr=null) {
        $response_json = NULL;
        try {

            $parameters = array(':version'   => $series->version,
                                ':signature' => $series->signature,
                                ':sigsha'   => hash('sha256',$series->signature),
                                ':t0'        => $series->t0,
                                ':t1'        => $series->t1,
                                ':domain'    => $this->_domain);
            $table = Config::config()::MySQL_Table($series->version);

            $query = "SELECT slice, unit FROM {$table}
                      WHERE 
                        version = :version 
                        AND signature in ( :signature , :sigsha ) 
                        AND ( 
                                (:t0 >= t0 AND :t0 <= t1) 
                                OR (:t0 <= t0 AND :t1 >= t1) 
                                OR (:t1 >= t0 AND :t1 <= t1 ) 
                            )";

            if($series->unit != null){
                $parameters = array_merge($parameters, array(':unit' => $series->unit));
                $query .= ' AND unit = :unit';
            }

            $invalid_sphere = (int)($series->x === null) +
                              (int)($series->y === null) +
                              (int)($series->z === null) +
                              (int)($series->r === null);

            if ($invalid_sphere != 0 && $invalid_sphere != 4) {
                throw new BadRequestException("Invalid sphere parameters");
            } else if($invalid_sphere == 0) {
                $parameters = array_merge($parameters, array(':x'    => $series->x,
                                                             ':y'    => $series->y,
                                                             ':z'    => $series->z,
                                                             ':r'    => $series->r));
                $query .= ' AND SQRT(POW(x - :x, 2) + POW(y - :y, 2) + POW(z - :z, 2)) <= :r + r';
                $filter_sphere = true;
            } else {
                $filter_sphere = false;
            }

            $query .= ' AND domain = :domain AND count > 0;';

            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME,
                                        Config::config()::MYSQL_PORT,
                                        Config::config()::MYSQL_DBNAME,
                                        Config::config()::MYSQL_USERNAME,
                                        Config::config()::MYSQL_PASSWORD);

            $select_stmt = $conn->prepare($query);
            $select_stmt->execute($parameters);

            $id_set = \Cassandra\Type::collection(\Cassandra\Type::bigint())->create();

            $selected_slices = array_column($select_stmt->fetchAll(), 'unit', 'slice');

            foreach ($selected_slices as $slice => $unit)
                $id_set->add(new \Cassandra\Bigint($slice));

            $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, $this->_username, $this->_password, Config::config()::CASSANDRA_KEYSPACE);

            $is_digital  = $series->unit()->is_digital();

            $cassandra_table = "{$this->_domain}" . (($is_digital) ? '_dig' : '');

            $points_limit = Config::config()::POINTS_LIMIT;
            $cass_select_stmt = $session->prepare("SELECT * FROM {$cassandra_table} WHERE (id IN ?) AND (timestamp >= ?) AND (timestamp <= ?) limit {$points_limit};");

            $result_page = $session->execute($cass_select_stmt,
                            array('arguments' => array($id_set, new \Cassandra\Varint($series->t0), new \Cassandra\Varint($series->t1))));

            //ini_set('memory_limit', '-1'); // check this
            //ini_set('max_execution_time', 300); // check this

            $index = 0;
            $response_json = array('series' => array());

            $aggregator = ($this->internal()) ? null : Aggregator::new($aggr);

            while($result_page) {
                foreach ($result_page as $column) {
                    $slice = (int)$column['id'];
                    $entry = $column['entry'];
                    $time  = (string)$column['timestamp']->value(); //bigint
                    if($is_digital){
                        $value_blob = $entry->get('value')->toBinaryString();
                        if(isset($_GET['type'])){
                            $value = Unpack::uInt64($value_blob);// blob
                        }else{
                            $value = Unpack::d64($value_blob);// blob
                        }
                        if($value === null)
                            $value = $value_blob; // TODO: remove this in the new API
                    } else {
                        $value = (double)$entry->get('value'); // double
                    }

                    $error = (int)$entry->get('error');             //tinyint
                    $conf  = (int)$entry->get('confidence');        //tinyint
                    $x     = (int)$entry->get('x');                 //int
                    $y     = (int)$entry->get('y');                 //int
                    $z     = (int)$entry->get('z');                 //int
                    $d     = (int)$entry->get('d');                 //int
                    $gw    = (int)$entry->get('gw');                //int (TO DO GET GATEWAY)

                    if ( $filter_sphere && !$series->contains_point($x, $y, $z) )
                        continue;

                    $smartdata = null;

                    switch ($series->version) {
                        case SmartData::STATIC_VERSION:
                            $smartdata = new StaticSmartData($series->unit,$value,$error,$conf,$x,$y,$z,$time,$d,$gw);
                            break;
                        case SmartData::MOBILE_VERSION:
                            $smartdata = new MobileSmartData($series->unit,$value,$error,$conf,$x,$y,$z,$time,$d,$gw, null);
                            break;
                        default:
                            continue 2;
                            break;
                    }

                    if($aggregator != null) {
                        $smartdata = $aggregator->aggregate($smartdata);
                    }

                    if($smartdata != null) {
                        if(is_array($smartdata)){
                            foreach ($smartdata as $key => $data) {
                                $response_json['series'][$index++] = $data->toArray();
                                if ($index >= Config::config()::POINTS_LIMIT) break;
                            }
                        } else {
                            $response_json['series'][$index++] = $smartdata->toArray();
                            if($this->internal()){
                                $response_json['series'][$index-1]['slice'] = $slice;
                                $response_json['series'][$index-1]['gw'] = $gw;
                            }
                        }
                    } else {
                        continue;
                    }

                    if ($index >= Config::config()::POINTS_LIMIT) break 2;
                }

                $result_page = $result_page->nextPage();

                if(!$result_page && isset($aggregator) && $index < Config::config()::POINTS_LIMIT){
                    $smartdata = $aggregator->finish();
                    if($smartdata != null) {
                        $response_json['series'][$index++] = $smartdata->toArray();
                    }
                }
            }

            $session->close();

            usort($response_json['series'], function ($a, $b) {
                $a_timestamp  = gmp_init($a['timestamp']);
                $b_timestamp  = gmp_init($b['timestamp']);
                return gmp_cmp($a_timestamp, $b_timestamp);
            });

        } catch (\Exception $e) {
            Logger::exception($e);
            $response_json = NULL;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }
        return $response_json;
    }

    public function insertBatch(Series $series, SmartData ...$smart_data) {
        if ($smart_data[0] instanceof StaticSmartData)
            return $this->_insertStaticSmartDataBatch($series, ...$smart_data);
        else
            throw new BadRequestException("Error processing JSON request: unsupported SmartData version");
    }

    // Insert a static version of smartdata
    private function _insertStaticSmartDataBatch(Series $series, StaticSmartData ...$smart_data) : bool {
        $sucess = true;

        $smart_data_count = count($smart_data);

        $parameters = array(':version'  => $series->version,
                            ':unit'     => $series->unit,
                            ':x'        => $series->x,
                            ':y'        => $series->y,
                            ':z'        => $series->z,
                            ':r'        => $series->r,
                            ':t0'       => $series->t0,
                            ':t1'       => $series->t1,
                            ':domain'   => $this->_domain);

        try {

            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME,
                                        Config::config()::MYSQL_PORT,
                                        Config::config()::MYSQL_DBNAME,
                                        Config::config()::MYSQL_USERNAME,
                                        Config::config()::MYSQL_PASSWORD);

            $table = Config::config()::MySQL_Table($series->version);

            $series_db = NULL;

            $stmt_select = $conn->prepare("SELECT * FROM {$table} WHERE version = :version
                                                    AND unit = :unit AND x = :x AND y = :y AND z = :z AND r = :r
                                                    AND t0 = :t0 AND t1 = :t1 AND domain = :domain
                                                    ORDER BY id DESC LIMIT 1;");

            $stmt_select->execute($parameters);
            $series_db = $stmt_select->fetch(\PDO::FETCH_OBJ);

            if ($series_db == NULL){// || (($series_db->count + $smart_data_count) >= Config\CASSANDRA_MAX_ROW_SIZE) ) {
                throw new Exception\NoMatchingSeriesException('No matching series');
            }else{

                $is_digital = $series->unit()->is_digital();

                $entry_type = \Cassandra\Type::userType(
                    'value',      ($is_digital) ? \Cassandra\Type::blob() : \Cassandra\Type::double(),
                    'error',      \Cassandra\Type::tinyint(),
                    'confidence', \Cassandra\Type::tinyint(),
                    'x',          \Cassandra\Type::int(),
                    'y',          \Cassandra\Type::int(),
                    'z',          \Cassandra\Type::int(),
                    'd',          \Cassandra\Type::bigint(),
                    'gw',         \Cassandra\Type::bigint()
                );

                $cassandra_table = "{$this->_domain}" . (($is_digital) ? '_dig' : '');

                /* change this */
                // TODO: get this values from the smartdata object
                $slice      = new \Cassandra\Bigint($series_db->slice);
                $error      = new \Cassandra\Tinyint(0);
                $confidence = new \Cassandra\Tinyint(0);
                $gw         = new \Cassandra\Bigint(-1);
                /* change this */


                $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, $this->_username, $this->_password, Config::config()::CASSANDRA_KEYSPACE);
                $insert_stmt = $session->prepare("INSERT INTO {$cassandra_table} (id, timestamp, entry) VALUES (?, ?, ?);");

                $batch = new \Cassandra\BatchStatement(\Cassandra::BATCH_LOGGED);

                $count = 0;
                foreach ($smart_data as $key => $data) {
                    $batch->add($insert_stmt, array($slice,
                                                    new \Cassandra\Varint($data->time),
                                                    $entry_type->create(
                                                        'value',      ($is_digital) ? new \Cassandra\Blob($data->value) : (double) $data->value,
                                                        'error',      $error,
                                                        'confidence', $confidence,
                                                        'x',          (int)$data->x,
                                                        'y',          (int)$data->y,
                                                        'z',          (int)$data->z,
                                                        'd',          new \Cassandra\Bigint($data->dev),
                                                        'gw',         $gw
                                                    )));
                    $count++;

                    if($count == 400){
                        //echo "       Inserted\n";
                        $session->execute($batch);
                        $count = 0;
                        $batch = new \Cassandra\BatchStatement(\Cassandra::BATCH_LOGGED);
                    }
                }

                if($count > 0){
                    $session->execute($batch);
                    $count = 0;
                }

                $update_stmt = $conn->prepare("UPDATE {$table} SET count = count + :smart_data_count WHERE id = :id LIMIT 1;");
                $update_stmt->execute(array('id' => $series_db->id, 'smart_data_count' => $smart_data_count));
            }

        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }

        return $sucess;
    }

    private function _insertMobileSmartDataBatch(Series $series, MobileSmartData ...$smart_data) : bool {
        throw new Exception\NotImplementedException("Batch insertion has not been implemented yet");
    }

    public function assignConfidence($t0, $t1, $max_time_diff, $query_filename="nn_input.csv", $confidence_filename="nn_output.csv", $train=false, $get=false) : bool {
        return true;
    }

    public function assign_confidence($smart_data) {
        $pushd = 'pushd '.Config::config()::CONFIDENCE_ASSIGNER;
        $popd  = 'popd';
        $json = $smart_data->toJson();
        //$return = shell_exec("{$pushd}; ./assign '{$json}'; {$popd};");
        preg_match('#\{(.*?)\}#', $return??'', $match);
        $confidence = $match[1] ?? $smart_data->confidence;
        return $smart_data->confidence = $confidence;
    }
}

## Backend of API version 1.1 ##############################################################################
class Backend_V1_1 extends Backend_Common
{
    public function create($to_be_created) {
        switch($to_be_created->version) {
        case SmartData::STATIC_VERSION:
            self::debug_X("Creating Series : {$to_be_created}");
            return $this->_createSeries($to_be_created);
        case SmartData::MOBILE_VERSION:
            self::debug_X("Creating Tracker : {$to_be_created}");
            return $this->_createTracker($to_be_created);
        default:
            throw new BadRequestException("Error processing request: unsupported Series/Tracker version [{$to_be_attached->version}]");
        }
    }

    public function attach($to_be_attached) {
        switch($to_be_attached->version) {
        case SmartData::STATIC_VERSION:
            self::debug_X("Attaching Series : {$to_be_attached}");
            return $this->_attachSeries($to_be_attached);
        case SmartData::MOBILE_VERSION:
            self::debug_X("Attaching Tracker : {$to_be_attached}");
            return $this->_attachTracker($to_be_attached);
        default:
            throw new BadRequestException("Error processing request: unsupported Series/Tracker version [{$to_be_attached->version}]");
            return null;
        }
    }

    // Insert SmartData into series/tracker matched
    public function insert(SmartData $smart_data) {
        if ($smart_data instanceof StaticSmartData) {
            self::debug_X("Inserting a Static SmartData : {$smart_data}");
            return $this->_insertStaticSmartData($smart_data);
        } else if ($smart_data instanceof MobileSmartData){
            self::debug_X("Inserting a Mobile SmartData : {$smart_data}");
            return $this->_insertMobileSmartData($smart_data);
        } else
            throw new BadRequestException("Error processing request: unsupported SmartData version");
    }

    public function insertBatch(...$batch) {
        if ($batch[0] instanceof Series){
            self::debug_X('Batching Series : ['.count($batch).']');
            return $this->_insertStaticSmartDataBatch(...$batch);
        } else if ($batch[0] instanceof Tracker){
            self::debug_X('Batching Tracker : ['.count($batch).']');
            return $this->_insertMobileSmartDataBatch(...$batch);
        } else
            throw new BadRequestException("Error processing request: unsupported data version");
    }


    public function insert_tracker_blob(MultiSmartData $msd) {
        $sucess = false;
        if ($msd->blob == null || strlen($msd->blob) == 0) {
            return false;
        }
        try {
            $min = $msd->time;
            $bytes = 4;
            if ($msd->encoding == 'I64' || $msd->encoding == 'D64') {
                $bytes = 8;
            }
            $num = strlen($msd->blob) / $bytes;
            $max = (int) ($msd->time + ($num * $msd->period));
            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME,
                                        Config::config()::MYSQL_PORT,
                                        Config::config()::MYSQL_DBNAME,
                                        Config::config()::MYSQL_USERNAME,
                                        Config::config()::MYSQL_PASSWORD);
            $table = Config::config()::MySQL_Table($msd->version);
            $series_db = NULL;

            $parameters = array(':version'   => $msd->version,
                                ':signature' => $msd->signature,
                                ':unit'      => $msd->unit->cod,
                                ':t0'        => $min,
                                ':t1'        => $max,
                                ':domain'    => $this->_domain,
                                ':dev'       => $msd->dev,
                                ':type'      => $msd->type,
                                ':period'    => $msd->period);

            $query = "SELECT * FROM {$table} WHERE version = :version
                        AND signature = :signature
                        AND unit = :unit
                        AND t0 <= :t0 AND t1 >= :t1
                        AND domain = :domain
                        AND dev = :dev
                        AND type = :type
                        AND period = :period
                        ORDER BY t0 LIMIT 1;";
            $stmt_select = $conn->prepare( $query );
            $stmt_select->execute($parameters);
            $sdb = $stmt_select->fetch(\PDO::FETCH_OBJ);
            if ($sdb != null && property_exists( $sdb, "version" ) ) {
                $trkr = new Tracker( $sdb->version, $sdb->unit, $sdb->signature, $sdb->t0, $sdb->t1, $sdb->domain, 
                                       $sdb->dev, $sdb->workflow, $sdb->type, $sdb->period, $sdb->uncertainty, $sdb->id, 
                                       $sdb->row_id, $sdb->event);
                $sucess = $this->_insertTrackerSmartDataBlob($trkr, $msd);
            } else {
                Logger::debug("Tracker not found!");
            }
        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }
        return $sucess;
    }

    public function insert_blob(MultiSmartData $msd) {
        $sucess = false;
        if ($msd->blob == null || strlen($msd->blob) == 0) {
            return false;
        }
        try {
            $min = $msd->time;
            $bytes = 4;
            if ($msd->encoding == 'I64' || $msd->encoding == 'D64') {
                $bytes = 8;
            }
            $num = strlen($msd->blob) / $bytes;
            $max = $msd->time + ($num * $msd->period);
            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME,
                                        Config::config()::MYSQL_PORT,
                                        Config::config()::MYSQL_DBNAME,
                                        Config::config()::MYSQL_USERNAME,
                                        Config::config()::MYSQL_PASSWORD);
            $table = Config::config()::MySQL_Table($msd->version);
            $series_db = NULL;

            $parameters = array(':version'   => $msd->version,
                                ':unit'      => $msd->unit->cod,
                                ':x'         => $msd->x,
                                ':y'         => $msd->y,
                                ':z'         => $msd->z,
                                ':t0'        => $min,
                                ':t1'        => $max,
                                ':type'      => $msd->type,
                                ':period'    => $msd->period,
                                ':domain'    => $this->_domain);

            $query = "SELECT * FROM {$table} WHERE version = :version
                        AND unit = :unit
                        AND r >= (SQRT(POW(x - :x, 2) + POW(y - :y, 2) + POW(z - :z, 2)))
                        AND t0 <= :t0 AND t1 >= :t1
                        AND domain = :domain
                        AND type = :type
                        AND period = :period
                        ORDER BY r ASC LIMIT 1;";
            $stmt_select = $conn->prepare( $query );
            $stmt_select->execute($parameters);
            $sdb = $stmt_select->fetch(\PDO::FETCH_OBJ);
            if ( $sdb != null && property_exists( $sdb, "version" ) ) {
                $series = new Series( $sdb->version, $sdb->unit, $sdb->x, $sdb->y, $sdb->z, $sdb->r, $sdb->t0, $sdb->t1, $msd->dev, $sdb->type, $sdb->period);
                $series->id = $sdb->id;
                $series->workflow = $sdb->workflow;
                $sucess = $this->_insertStaticSmartDataBlob($series, $msd);
            } 
        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }
        return $sucess;
    }
    
    public function insert_multiple(SmartData ...$smart_data) {
        $sucess = true;
        $smart_data_count = count($smart_data);

        if ($smart_data_count < 1) {
            return;
        }

        // First: define max and min timestamp, and verify if a there is a series that contains all

        $sd =  $smart_data[0];
        $min = $sd->time;
        $max = $min;
        foreach ($smart_data as $sd) {
            if ($sd->time > $max) {
                $max = $sd->time;
            } else if ($sd->time < $min) {
                $min = $sd->time; 
            }
        }

        try {

            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME,
                                        Config::config()::MYSQL_PORT,
                                        Config::config()::MYSQL_DBNAME,
                                        Config::config()::MYSQL_USERNAME,
                                        Config::config()::MYSQL_PASSWORD);

            $table = Config::config()::MySQL_Table($sd->version);
            $series_db = NULL;

            $parameters = array(':version'   => $sd->version,
                                ':unit'      => $sd->unit,
                                ':x'         => $sd->x,
                                ':y'         => $sd->y,
                                ':z'         => $sd->z,
                                ':t0'        => $min,
                                ':t1'        => $max,
                                ':domain'    => $this->_domain);

            $query = "SELECT * FROM {$table} WHERE version = :version
                        AND unit = :unit
                        AND r >= (SQRT(POW(x - :x, 2) + POW(y - :y, 2) + POW(z - :z, 2)))
                        AND t0 <= :t0 AND t1 >= :t1
                        AND domain = :domain
                        ORDER BY r ASC LIMIT 1;";

            $stmt_select = $conn->prepare( $query );
            $stmt_select->execute($parameters);
            $sdb = $stmt_select->fetch(\PDO::FETCH_OBJ);
            if ($sdb != null && property_exists( $sdb, "version" ) ) {
                $series = new Series( $sdb->version, $sdb->unit, $sdb->x, $sdb->y, $sdb->z, $sdb->r, $sdb->t0, $sdb->t1, $sd->dev);
                $series->id = $sdb->id;
                $sucess = $this->_insertStaticSmartDataBatch($series, ...$smart_data);
            } else {
                throw new BadRequestException("Series was not found, check sent data.");
            }
       } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }

    return $sucess;
    }


    // Create a new series if not exist
    private function _createSeries(Series $series) : bool {
        $sucess = true;
        
        $sel_parameters = array(':version'  => $series->version,
                            ':unit'     => $series->unit,
                            ':x'        => $series->x,
                            ':y'        => $series->y,
                            ':z'        => $series->z,
                            ':r'        => $series->r,
                            ':period'   => $series->period,
                            ':workflow' => $series->workflow ?? 0,
                            ':domain'   => $this->_domain);
                            
        $parameters = array(':version'  => $series->version,
                            ':unit'     => $series->unit,
                            ':x'        => $series->x,
                            ':y'        => $series->y,
                            ':z'        => $series->z,
                            ':r'        => $series->r,
                            ':t0'       => $series->t0,
                            ':t1'       => $series->t1,
                            ':workflow' => $series->workflow ?? 0,
                            ':domain'   => $this->_domain,
                            ':type'     => $series->type,
                            ':period'   => $series->period,
                            ':event'    => $series->event,
                            ':count'    => $series->count,
                            ':uncertainty' => $series->uncertainty);
        try {
            Unit::validate( $series->unit );
            // As there can be OPEN series (with t0 = 0, waiting first event, or t1 = 99999999999999, wainting to be finished) NO new series should be created until they are closed.
            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, Config::config()::MYSQL_PASSWORD);
            $table = Config::config()::MySQL_Table($series->version);
            
            $stmt_select = $conn->prepare("SELECT * FROM {$table} WHERE version = :version AND unit = :unit AND (SQRT(POW(x - :x, 2) + POW(y - :y, 2) + POW(z - :z, 2))) <= (r + :r) AND (t0 = 0 OR t1 = 9999999999999999) AND workflow = :workflow AND domain = :domain AND period = :period");
            $stmt_select->execute($sel_parameters);
            $aux = $stmt_select->fetch(\PDO::FETCH_OBJ);
            if ($aux != NULL) {
                throw new Exception\InternalException('There are open series that must be finished first.');
            }
            
            $sel_parameters[':t0'] = $series->t0;
            $sel_parameters[':t1'] = $series->t1;
            $stmt_select = $conn->prepare("SELECT * FROM {$table} WHERE version = :version AND unit = :unit AND (SQRT(POW(x - :x, 2) + POW(y - :y, 2) + POW(z - :z, 2))) <= (r + :r) AND ((:t0 >= t0 AND :t0 <= t1) OR (:t0 <= t0 AND :t1 >= t1) OR (:t1 >= t0 AND :t1 <= t1 )) AND workflow = :workflow AND domain = :domain AND period = :period");

            $stmt_select->execute($sel_parameters);
            $series_set = $stmt_select->fetch(\PDO::FETCH_OBJ);

            if ($series_set == NULL) {
                $stmt_insert = $conn->prepare("INSERT INTO {$table} VALUES (DEFAULT, :version, :unit, :x, :y, :z, :r, DEFAULT, :t0, :t1, :workflow, :domain, DEFAULT, :type, :period, :event, :count, :uncertainty);");
                $stmt_insert->execute($parameters);
                self::debug_X('The Series was created!');
            } else {
                // when only one series exists, does not return an array, but a stdClass 
                if ( ! is_array( $series_set) ) {
                     $aux_s = $series_set;
                     $series_set = array();
                     array_push($series_set, $aux_s);
                }
                $rows_table = '`rows`';
                foreach ($series_set as $series_aux) {
                    $dist = sqrt(pow($series_aux->x - $series->x, 2) + pow($series_aux->y - $series->y, 2) + pow($series_aux->z - $series->z, 2));
                    $t0 = ($series_aux->t0 <= $series->t0) ? $series_aux->t0 : $series->t0;
                    $t1 = ($series_aux->t1 >= $series->t1) ? $series_aux->t1 : $series->t1;
                    if ($series_aux->r >= ($dist + $series->r)) {
                        //new series is spatially contained by the already inserted one
                        $x = $series_aux->x;
                        $y = $series_aux->y;
                        $z = $series_aux->z;
                        $r = $series_aux->r;
                    } elseif ($series->r >= ($dist + $series_aux->r)) {
                        //already inserted series is spatially contained by the new one
                        $x = $series->x;
                        $y = $series->y;
                        $z = $series->z;
                        $r = $series->r;
                    } else {
                        $r = ($series->r + $dist + $series_aux->r)/2;
                        $x = ($series_aux->x + (($series_aux->x - $series->x) * ($r - $series_aux->r)) / $dist);
                        $y = ($series_aux->y + (($series_aux->y - $series->y) * ($r - $series_aux->r)) / $dist);
                        $z = ($series_aux->z + (($series_aux->z - $series->z) * ($r - $series_aux->r)) / $dist);
                        //http://answers.google.com/answers/threadview/id/342125.html
                    }
                    if (isset($series->id) && $t1 > 0) {
                        $parameters = array(':x'        => $x,
                            ':y'        => $y,
                            ':z'        => $z,
                            ':r'        => $r,
                            ':t0'       => $t0,
                            ':t1'       => $t1,
                            ':id'       => $series->id);
                        $update_stmt = $conn->prepare("UPDATE {$table} SET x = :x, y = :y, z = :z, r = :r, t0 = :t0, t1 = :t1 WHERE id = :id LIMIT 1;");
                        $update_stmt->execute($parameters);
                        $parameters_up = array(':id' => $series->id,
                                               ':id_aux' => $series_aux->id);
                        $update_stmt = $conn->prepare("UPDATE {$rows_table} SET series_id = :id WHERE series_id = :id_aux LIMIT 1;");
                        $update_stmt->execute($parameters_up);
                        $parameters_del = array(':id' => $series_aux->id);
                        $update_stmt = $conn->prepare("DELETE FROM {$table} WHERE id = :id LIMIT 1;");
                        $update_stmt->execute($parameters_del);
                    } else if ($t1 > 0) {
                        $parameters = array(':x'        => $x,
                            ':y'        => $y,
                            ':z'        => $z,
                            ':r'        => $r,
                            ':t0'       => $t0,
                            ':t1'       => $t1,
                            ':id'       => $series_aux->id);
                        $update_stmt = $conn->prepare("UPDATE {$table} SET x = :x, y = :y, z = :z, r = :r, t0 = :t0, t1 = :t1 WHERE id = :id LIMIT 1;");
                        $update_stmt->execute($parameters);
                        $series->id = (int) $series_aux->id;
                    }
                    $series->x = (int) $x;
                    $series->y = (int) $y;
                    $series->z = (int) $z;
                    $series->r = (int) $r;
                    $series->t0 = (int) $t0;
                    $series->t1 = (int) $t1;
                }
                //self::debug_X("error");
            }

        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            $conn = null;
        }

        return $sucess;
    }

    // Create a new tracker if not exist
    private function _createTracker(Series $tracker) : bool {
        $sucess = true;

        $parameters = array(':version'   => $tracker->version,
                            ':unit'      => $tracker->unit,
                            ':signature' => $tracker->signature ?? 0,
                            ':sigsha'   => hash('sha256',$tracker->signature),
                            ':t0'        => $tracker->t0,
                            ':t1'        => $tracker->t1,
                            ':dev'       => $tracker->dev,
                            ':workflow'  => $tracker->workflow ?? 0,
                            ':domain'    => $this->_domain,
                            ':period'    => $tracker->period ?? 0,
                            ':type'      => $tracker->type ?? 'OLD');

        try {
            if (!property_exists($tracker,"signature") || $tracker->signature === 0) throw new Exception\InternalException('Signature must be provided in Mobile SmartData.');
            Unit::validate( $tracker->unit );
            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, Config::config()::MYSQL_PASSWORD);
            $table = Config::config()::MySQL_Table($tracker->version);
            $stmt_select = $conn->prepare("SELECT * FROM {$table} WHERE
                  version = :version
                  AND unit = :unit
                  AND dev = :dev
                  AND t0 = :t0
                  AND t1 = :t1
                  AND signature in ( :signature , :sigsha )
                  AND workflow = :workflow
                  AND domain = :domain
                  AND period = :period
                  AND type = :type
                ORDER BY count ASC LIMIT 1;");

            $stmt_select->execute($parameters);
            $selected_tracker = $stmt_select->fetch(\PDO::FETCH_OBJ);

            if ($selected_tracker == NULL) {
                self::debug_X('The Tracker did not exist before A');
                $parameters[':event'] = $trakcer->event ?? null; 
                $parameters[':uncertainty'] = $tracker->uncertainty ?? 0;
                
                $stmt_insert = $conn->prepare("INSERT INTO {$table} (id, version, unit, dev, t0, t1, signature, count, workflow, domain, row_id, updated_at, type, period, event, uncertainty) 
                                               VALUES (DEFAULT, :version, :unit, :dev, :t0, :t1, :sigsha, DEFAULT, :workflow, :domain, (SELECT counter(:domain)), DEFAULT, :type, :period, :event, :uncertainty);");
                $stmt_insert->execute($parameters);
                self::debug_X("Tracker sucessfully  created");
            } 
        } catch (\Exception $e) {
            Logger::debug("Exception: ".$e);
            throw $e;
        } finally {
            $conn = null;
        }

        return $sucess;
    }

    // Create a new series if there is not one that contains it
    private function _attachSeries(Series $series) : bool {
        $sucess = true;

        $parameters = array(':version'  => $series->version,
                            ':unit'     => $series->unit,
                            ':x'        => $series->x,
                            ':y'        => $series->y,
                            ':z'        => $series->z,
                            ':r'        => $series->r,
                            //':dev'      => $series->dev, // TODO: The series doesn't has a dev
                            ':t0'       => $series->t0,
                            ':t1'       => $series->t1,
                            ':workflow' => $series->workflow ?? 0,
                            ':domain'   => $this->_domain,
                            ':type'     => $series->type,
                            ':period'   => $series->period);

        try {
            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, Config::config()::MYSQL_PASSWORD);
            $table = Config::config()::MySQL_Table($series->version);
            $stmt_select = $conn->prepare("SELECT * FROM {$table} WHERE version = :version
                  AND unit = :unit
                  AND r >= (SQRT(POW(x - :x, 2) + POW(y - :y, 2) + POW(z - :z, 2)) + :r)
                  AND t0 <= :t0 AND t1 >= :t1
                  AND workflow = :workflow
                  AND domain = :domain
                  AND type = :type
                  AND period = :period
                ORDER BY r ASC LIMIT 1;");

            $stmt_select->execute($parameters);
            $series = $stmt_select->fetch(\PDO::FETCH_OBJ);

            if ($series == NULL) {
                $parameters = array(':version'  => $series->version,
                        ':unit'     => $series->unit,
                        ':x'        => $series->x,
                        ':y'        => $series->y,
                        ':z'        => $series->z,
                        ':r'        => $series->r,
                        ':t0'       => $series->t0,
                        ':t1'       => $series->t1,
                        ':workflow' => $series->workflow ?? 0,
                        ':domain'   => $this->_domain,
                        ':type'     => $series->type,
                        ':period'   => $series->period,
                        ':event'    => $series->event,
                        ':count'    => $series->count,
                        ':uncertainty' => $series->uncertainty);
                $stmt_insert = $conn->prepare("INSERT INTO {$table} VALUES (DEFAULT, :version, :unit, :x, :y, :z, :r, DEFAULT, :t0, :t1, :workflow, :domain, DEFAULT, :type, :period, :event, :count, :uncertainty);");
                $stmt_insert->execute($parameters);
                self::debug_X('The Series was created...');
            } else {
                //self::debug_X($series);
            }
        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            $conn = null;
        }

        return $sucess;
    }

    // Create a new series if there is not one that contains it
    private function _attachTracker(Series $tracker) : bool {
        $sucess = true;

        $parameters = array(':version'   => $tracker->version,
                            ':unit'      => $tracker->unit,
                            ':signature' => $tracker->signature,
                            ':sigsha'   => hash('sha256',$tracker->signature),
                            ':t0'        => $tracker->t0,
                            ':t1'        => $tracker->t1,
                            ':dev'       => $tracker->dev,
                            ':workflow'  => $tracker->workflow ?? 0,
                            ':domain'    => $this->_domain);

        try {

            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, Config::config()::MYSQL_PASSWORD);
            $table = Config::config()::MySQL_Table($tracker->version);
            $stmt_select = $conn->prepare("SELECT * FROM {$table} WHERE
                  version = :version
                  AND unit = :unit
                  AND dev = :dev
                  AND t0 <= :t0 AND t1 >= :t1
                  AND signature in ( :signature, :sigsha )
                  AND workflow = :workflow
                  AND domain = :domain
                ORDER BY count ASC LIMIT 1;");

            $stmt_select->execute($parameters);
            $selected_tracker = $stmt_select->fetch(\PDO::FETCH_OBJ);

            if ($selected_tracker == NULL) {
                self::debug_X('The Tracker did not exist before B');
                $stmt_insert = $conn->prepare("INSERT INTO {$table} VALUES (DEFAULT, :version, :unit, :dev, :t0, :t1, :sigsha, DEFAULT, :workflow, :domain, (SELECT counter(:domain)), DEFAULT);");
                $stmt_insert->execute($parameters);
            } else {
                //self::debug_X($selected_tracker);
            }

        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            $conn = null;
        }

        return $sucess;
    }

    // Insert a static version of smartdata
    private function _insertStaticSmartData(StaticSmartData $smart_data) : bool {
        $sucess = true;
        $parameters = array(':unit'       => $smart_data->unit,
                            ':x'          => $smart_data->x,
                            ':y'          => $smart_data->y,
                            ':z'          => $smart_data->z,
                            //':dev'        => $smart_data->dev, // TODO: The series doesn't has a dev
                            ':time'       => $smart_data->time,
                            ':domain'     => $this->_domain);

        $is_digital = $smart_data->unit()->is_digital();

        try {
            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, Config::config()::MYSQL_PASSWORD);
            $table = Config::config()::MySQL_Table($smart_data->version);
            $stmt_select = $conn->prepare("SELECT * FROM {$table} WHERE unit = :unit AND (SQRT(POW(x - :x, 2) + POW(y - :y, 2) + POW(z - :z, 2))) <= r AND :time >= t0 AND :time <= t1 AND domain = :domain and type <> 'TTH';");
            $stmt_select->execute($parameters);
            $series_set = $stmt_select->fetchAll(\PDO::FETCH_OBJ);
            if ($series_set == NULL || ( count($series_set) === 0 ) ) {
                throw new Exception\NoMatchingSeriesException('No matching series');
            } else {
                // the series exists
                $rows_table = '`rows`'; // TODO: move to Config file (row table name)

                $count_limit = Config::config()::CASSANDRA_MAX_ROW_SIZE;
                foreach ($series_set as $series) {
                    $series->row = NULL;
                    $first = true;

                    do {
                        $stmt_select_row = $conn->prepare("SELECT * FROM {$rows_table} WHERE series_id in (:series_id) AND x = :x AND y = :y AND z = :z AND dev = :dev AND count < {$count_limit} LIMIT 1;");

                        $parameters_row = array(':series_id'  => $series->id,
                                                ':x'          => $smart_data->x,
                                                ':y'          => $smart_data->y,
                                                ':z'          => $smart_data->z,
                                                ':dev'        => $smart_data->dev);

                        $stmt_select_row->execute($parameters_row);
                        $series->row = $stmt_select_row->fetch(\PDO::FETCH_OBJ);
                        if ($first && $series->row==NULL) {
                            self::debug_X('There were no rows before');
                            $stmt_insert_row = $conn->prepare("INSERT INTO {$rows_table} VALUES (DEFAULT, :x, :y, :z, :dev, DEFAULT, (SELECT counter(:domain)), :series_id, DEFAULT);");
                            $stmt_insert_row->execute(array_merge($parameters_row, array(':domain' => $this->_domain)));
                            $first = false;
                        } else {
                            break;
                        }
                    } while (true);
                }
            }
            $entry_type = \Cassandra\Type::userType(
                'value',      ($is_digital) ? \Cassandra\Type::blob() : \Cassandra\Type::double(),
                'error',      \Cassandra\Type::tinyint(),
                'confidence', \Cassandra\Type::tinyint(),
                'gw',         \Cassandra\Type::bigint()
            );

            foreach ($series_set as $series) {
                self::debug_X('Inserting StaticSmartData on series ['.$series->id.'] row ['.$series->row->row_id.']');
                $_smart_data = null;
                if($series->workflow ?? false){
                    self::debug_X('Enter workflow');
                    $_smart_data = clone $smart_data;
                    $_smart_data = $this->_workflow_input($_smart_data, $series->workflow);
                }else{
                    $_smart_data = $smart_data;
                }
                if($_smart_data === null)
                    continue;

                $arguments = array(
                    new \Cassandra\Bigint($series->row->row_id),
                    new \Cassandra\Varint($_smart_data->time),
                    $entry_type->create(
                        'value',      ($is_digital) ? new \Cassandra\Blob($_smart_data->value) : (double) $_smart_data->value,
                        'error',      new \Cassandra\Tinyint($_smart_data->error),
                        'confidence', new \Cassandra\Tinyint($_smart_data->confidence),
                        'gw',         new \Cassandra\Bigint($this->_gw_id)
                    )
                );

                // TODO: alter the cassandra table's name to Config file
                $cassandra_table  = $this->_domain.'_v1_1_';
                $cassandra_table .= $is_digital ? 'digital' : 'si';
                // -----

                $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, $this->_username, $this->_password, Config::config()::CASSANDRA_KEYSPACE);

                $insert_stmt = $session->prepare("INSERT INTO {$cassandra_table} (id, timestamp, entry) VALUES (?, ?, ?);");
                $session->execute($insert_stmt, array('arguments' => $arguments));

                $update_stmt = $conn->prepare("UPDATE {$rows_table} SET count = count + 1 WHERE id = :id LIMIT 1;");
                $update_stmt->execute(array('id' => $series->row->id));
                
                if ($series->t0 == 0) {
                    $upd_stmt = $conn->prepare("UPDATE {$table} SET t0 = :t0 where id = :id LIMIT 1;");
                    $instime = time() * 1000000;
                    $upd_stmt->execute(array('id' => $series->id, 't0' => $instime));
                }
            }
        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }
        return $sucess;
    }

    private function _process_exists($proc_id) {
        $exists= false;
        exec("ps -A | grep -i $proc_id | grep -v grep", $pids);
        if (!empty($pids)) {
            $exists = true;
        }
        return ($exists || file_exists("/proc/{$proc_id}"));
    }

    private function _initialize_daemon($workflow) {
        $_root  = __DIR__;
        //self::debug_X("root at: {$_root}");
        if (file_exists("{$_root}/../workflow/{$this->_domain}/{$workflow}_daemon")) {
            $alive = false;
            try {
                if(file_exists("{$_root}/../workflow/{$this->_domain}/{$workflow}_pid")) {
                    $pid_file = fopen("{$_root}/../workflow/{$this->_domain}/{$workflow}_pid", "r");
                    if($pid_file) {
                        $pid = fgets($pid_file);
                        fclose($pid_file);
                        $alive = $this->_process_exists($pid) ||  $this->_process_exists("{$workflow}_daemon"); //maybe this string check is not valid
                    }
                }
            } catch (Exception $e) {
                $alive = false;
            }

            if (! $alive) {
                if (file_exists("{$_root}/../workflow/{$this->_domain}/{$workflow}_error.txt"))
                    unlink("{$_root}/../workflow/{$this->_domain}/{$workflow}_error.txt");
                if (file_exists("{$_root}/../workflow/{$this->_domain}/{$workflow}_pid"))
                    unlink("{$_root}/../workflow/{$this->_domain}/{$workflow}_pid");
                if (file_exists("{$_root}/../workflow/{$this->_domain}/{$workflow}_input"))
                    unlink("{$_root}/../workflow/{$this->_domain}/{$workflow}_input");
                if (file_exists("{$_root}/../workflow/{$this->_domain}/{$workflow}_output"))
                    unlink("{$_root}/../workflow/{$this->_domain}/{$workflow}_output");

                $descriptorspec = array(
                   0 => array("pipe", "r"),  // stdin is a pipe that the child will read from
                   1 => array("pipe", "w"),  // stdout is a pipe that the child will write to
                   2 => array("file", "{$_root}/../workflow/{$this->_domain}/{$workflow}_error.txt", "a") // stderr is a file to write to
                );
                $path = "{$_root}/../workflow/{$this->_domain}";
                $pid = proc_open("./{$workflow}_daemon $> {$workflow}_results.log", $descriptorspec, $pipes, $path);

                $timeout = 100000; //what is the best value? we can use a timeout in seconds also
                while (!(file_exists("{$_root}/../workflow/{$this->_domain}/{$workflow}_pid")) && $timeout >= 0)
                    $timeout--;
                if($timeout < 0) {
                    self::debug_X("Fail to initialize daemon for workflow {$this->_domain}/{$workflow}");
                }
            }
        } else {
            self::debug_X("No daemon found for {$this->_domain}/{$workflow}");
        }
    }

    private function _notify(\stdClass $json) {
        if(isset($json->notify)) {
            $notification = "Domain: {$this->_domain}. Data irregularity.";
            if(isset($json->notify->description)) {
                $notification .= " Description: {$json->notify->description}.";
            }
            if(isset($json->notify->severity)) {
                //threshold can be either defined by the workflow or by a standard value (100% in this case)
                // isset($json->notify->severity_threshold)
                $notification .= " Severity level: {$json->notify->severity}.";
                if ($json->notify->severity > 100) {
                    //send mail or message bus
                }
            }
            //log notification
            self::debug_X($notification);
        }
    }

    private function _workflow_tth_input($msd, $wf) {
        $_root  = __DIR__;
        $_wf = "in{$wf}";
        $pushd = "pushd {$_root}/../workflow/{$this->_domain} > /dev/null";
        $popd  = 'popd > /dev/null';
        $time = intval(1000000 * microtime(true));
        $this->_initialize_daemon($_wf);
        $str_json = $msd->toJson();
        $filename = "{$msd->unit_code}_{$msd->time}_{$msd->x}_{$msd->y}_{$msd->z}_{$time}";
        $input = fopen("{$_root}/../workflow/{$this->_domain}/{$_wf}_{$filename}", "w");
        fwrite($input, $str_json."\n");
        fclose($input);
        $_REQUEST['stats']['before_shell_call'] =  intval(1000000 * microtime(true));
        $return = shell_exec("/bin/bash -c ' {$pushd}; ./{$_wf} {$filename}; {$popd}; '");
        $_REQUEST['stats']['before_return_decode'] =  intval(1000000 * microtime(true));
        $json = json_decode($return, false, 512, JSON_BIGINT_AS_STRING);
        unlink("{$_root}/../workflow/{$this->_domain}/{$_wf}_{$filename}");
        $error = json_last_error();
        if ($error === JSON_ERROR_NONE){
            $smart_data = MultiSmartData::multi_values($json->series);
        }else{
            throw new Exception\InternalException('Error while processing workflow!');
        }
        $this->_notify($json);
        return $smart_data;
    }

    private function _workflow_input($smart_data, $wf) {
        $_root  = __DIR__;
        $_wf = "in{$wf}";
        $pushd = "pushd {$_root}/../workflow/{$this->_domain} > /dev/null";
        $popd  = 'popd > /dev/null';
        $this->_initialize_daemon($_wf);
        $compr = strlen($smart_data->value);
        $js = $smart_data->toJson();
        if ($compr > 500) {
            $time = intval(1000000 * microtime(true));
            $filename = "{$smart_data->unit_code}_{$smart_data->time}_{$smart_data->x}_{$smart_data->y}_{$smart_data->z}_{$time}";
            $input = fopen("{$_root}/../workflow/{$this->_domain}/{$_wf}_".$filename, "w");
            fwrite($input, "{\"smartdata\":" . $js . "}\n");
            fclose($input);
            $return = shell_exec("/bin/bash -c ' {$pushd}; ./{$_wf} {$filename}; {$popd}; '");
            unlink("{$_root}/../workflow/{$this->_domain}/{$_wf}_".$filename);
        } else {
            $json = "{\"smartdata\":" . $js . "}";
            $return = shell_exec("{$pushd}; ./{$_wf} '{$json}'; {$popd};");
        }
        $json = json_decode($return, false, 512, JSON_BIGINT_AS_STRING);
        $error = json_last_error();
        if ($error === JSON_ERROR_NONE){
            $smart_data = SmartData::fromJson($json->smartdata);
        }else{
            throw new Exception\InternalException('Error while processing workflow!');
        }
        $this->_notify($json);
        return $smart_data;
    }

    private function _workflow_output($json, $wf, $series) {
        $_root  = __DIR__;
        $_wf = "out{$wf}";
        $pushd = "pushd {$_root}/../workflow/{$this->_domain} > /dev/null";
        $popd  = "popd > /dev/null";
        $this->_initialize_daemon($_wf);
        $u = Unit::interpret($series->unit);
        if ($u->is_digital() && ( $series->unit & 0x0FFF0000 ) == 0x02230000 ) {
            $aux = base64_encode($json['series'][0]['value']);
            $json['series'][0]['value'] = $aux;
        }
        $str_json = json_encode($json);
        $time = intval(1000000 * microtime(true));
        $filename = "{$series->unit}_{$series->t0}_{$series->t1}_{$series->x}_{$series->y}_{$series->z}_{$series->r}_{$time}";
        $input = fopen("{$_root}/../workflow/{$this->_domain}/{$_wf}_{$filename}", "w");
        fwrite($input, "$str_json\n");
        fclose($input);
        $return = shell_exec("/bin/bash -c ' {$pushd}; ./{$_wf} {$filename}; {$popd}; '");
        $json = json_decode($return, true, 512, JSON_BIGINT_AS_STRING);
        $err_json = json_last_error();
        if ($err_json !== JSON_ERROR_NONE){
            self::debug_X("JSON = {$return}, error = {$err_json}");
            throw new Exception\InternalException('Error while processing workflow!');
        }
        unlink("{$_root}/../workflow/{$this->_domain}/{$_wf}_{$filename}");
        return $json;
    }

    // Insert a mobile version of smartdata
    private function _insertMobileSmartData(MobileSmartData $smart_data) : bool {
        // TODO: check if it is working fine
        $sucess = true;
        $parameters = array(':version'    => $smart_data->version,
                            ':unit'       => $smart_data->unit,
                            ':signature'  => $smart_data->signature,
                            ':sigsha'   => hash('sha256',$smart_data->signature),
                            ':time'       => $smart_data->time,
                            ':dev'        => $smart_data->dev,
                            ':domain'     => $this->_domain);
        $is_digital = $smart_data->unit()->is_digital();
        try {
            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, Config::config()::MYSQL_PASSWORD);
            $table_trackers = Config::config()::MySQL_Table($smart_data->version);
            $tracker_set = NULL;
            $first = true;
            do {
                $sql="SELECT t.* FROM (SELECT *, min(count) as min FROM trackers WHERE version = :version AND unit = :unit AND signature in ( :signature , :sigsha ) AND :time >= t0 AND :time <= t1 AND dev = :dev AND domain = :domain GROUP BY version, unit, dev, t0, t1, signature, workflow, domain) as x inner join trackers as t on t.version = x.version AND t.unit = x.unit AND t.dev = x.dev AND t.t0 = x.t0 AND t.t1 = x.t1 AND t.signature = x.signature AND t.workflow = x.workflow AND t.domain = x.domain AND t.count = x.min GROUP BY version, unit, dev, t0, t1, signature, workflow, domain;";
                $stmt_select = $conn->prepare($sql);
                $stmt_select->execute($parameters);
                $tracker_set = $stmt_select->fetchAll(\PDO::FETCH_OBJ);
                if ($tracker_set == NULL) {
                    // TODO: change do NoMatchingTrackerException
                    Logger::debug("No MATCHIN TRACKER!!!");
                    throw new Exception\NoMatchingSeriesException('No matching tracker');
                } else {
                    if ($first){
                        foreach ($tracker_set as $tracker) {
                            if($tracker->count>=Config::config()::CASSANDRA_MAX_ROW_SIZE){
                                $stmt_insert = $conn->prepare("INSERT INTO {$table_trackers} VALUES (DEFAULT, :version, :unit, :dev, :t0, :t1, :sigsha, DEFAULT, :workflow, :domain, (SELECT counter(:domain)), DEFAULT);");
                                $stmt_insert->execute(array(':version'   => $tracker->version,
                                                            ':unit'      => $tracker->unit,
                                                            ':signature' => $tracker->signature,
                                                            ':sigsha'    => hash('sha256',$smart_data->signature),
                                                            ':t0'        => $tracker->t0,
                                                            ':t1'        => $tracker->t1,
                                                            ':dev'       => $tracker->dev,
                                                            ':workflow'  => $tracker->workflow,
                                                            ':domain'    => $this->_domain));
                            }
                        }
                        $first = false;
                    } else {
                        break;
                    }
                }
            } while (true);

            $entry_type = \Cassandra\Type::userType(
                'value',      ($is_digital) ? \Cassandra\Type::blob() : \Cassandra\Type::double(),
                'error',      \Cassandra\Type::tinyint(),
                'confidence', \Cassandra\Type::tinyint(),
                'x',          \Cassandra\Type::int(),
                'y',          \Cassandra\Type::int(),
                'z',          \Cassandra\Type::int(),
                'gw',         \Cassandra\Type::bigint()
            );
            $inside = 0;
            foreach ($tracker_set as $tracker) {
                self::debug_X('Inserting MobileSmartData on Tracker ['.$tracker->id.'] row ['.$tracker->row_id.']');
                if($tracker->workflow ?? false){
                    self::debug_X('Entering V1.2 workflow '. $tracker->workflow);
                    $_smart_data = clone $smart_data;
                    $smart_data = $this->_workflow_input($_smart_data, $tracker->workflow);
                    if ($_smart_data === null) {
                        continue;
                    }
                }
                $arguments = array(
                    new \Cassandra\Bigint($tracker->row_id),
                    new \Cassandra\Varint($smart_data->time),
                    $entry_type->create(
                        'value',      ($is_digital) ? new \Cassandra\Blob($smart_data->value) : (double) $smart_data->value,
                        'error',      new \Cassandra\Tinyint($smart_data->error),
                        'confidence', new \Cassandra\Tinyint($smart_data->confidence),
                        'x',          (int)$smart_data->x,
                        'y',          (int)$smart_data->y,
                        'z',          (int)$smart_data->z,
                        'gw',         new \Cassandra\Bigint($this->_gw_id)
                    )
                );

                // TODO: alter the cassandra table's name to Config file
                $cassandra_table  = $this->_domain.'_v1_2_';
                $cassandra_table .= $is_digital ? 'digital' : 'si';
                
                $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, $this->_username, $this->_password, Config::config()::CASSANDRA_KEYSPACE);

                $insert_stmt = $session->prepare("INSERT INTO {$cassandra_table} (id, timestamp, entry) VALUES (?, ?, ?);");
                $session->execute($insert_stmt, array('arguments' => $arguments));

                $update_stmt = $conn->prepare("UPDATE {$table_trackers} SET count = count + 1 WHERE id = :id LIMIT 1;");
                $update_stmt->execute(array('id' => $tracker->id));
                self::debug_X(" inserted at cassandra! ");
                $inside = $inside + 1 ;
            }
            if ($inside == 0) {
                self::debug_X("NÃO ENTROU!!!!");
            }
        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }

        return $sucess;
    }

    private function _insertStaticSmartDataBatch(Series $series, StaticSmartData ...$smart_data) : bool {
        if($series->r != 0) {
            throw new Exception\NoMatchingSeriesException('Insertion series must have r = 0, found'.$series->r ); // TODO: change this exception type
        }
        $sucess = true;
        $smart_data_count = count($smart_data);

        $parameters = array(':version'  => $series->version,
                            ':unit'     => $series->unit,
                            ':x'        => $series->x,
                            ':y'        => $series->y,
                            ':z'        => $series->z,
                            ':r'        => $series->r,
                            ':t0'       => $series->t0,
                            ':t1'       => $series->t1,
                            ':type'     => $series->type,
                            ':period'   => $series->period,
                            ':domain'   => $this->_domain);
        try {
            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, Config::config()::MYSQL_PASSWORD);
            $table = Config::config()::MySQL_Table($series->version);
            $series_db = NULL;
            $stmt_select = $conn->prepare("SELECT * FROM {$table} WHERE version = :version AND unit = :unit AND x = :x AND y = :y AND z = :z AND r = :r AND t0 = :t0 AND t1 = :t1 AND domain = :domain AND type = :type AND period = :period LIMIT 1;");
            $stmt_select->execute($parameters);
            $series_db = $stmt_select->fetch(\PDO::FETCH_OBJ);

            if ($series_db == NULL){// || (($series_db->count + $smart_data_count) >= Config\CASSANDRA_MAX_ROW_SIZE) ) {
                throw new Exception\NoMatchingSeriesException('No matching series');
            }else{
                $rows_table = '`rows`'; // TODO: move to Config file (row table name)
                $count_limit = Config::config()::CASSANDRA_MAX_ROW_SIZE;
                $series_db->row = NULL;
                $first = true;
                do {
                    $stmt_select_row = $conn->prepare("SELECT * FROM {$rows_table} WHERE series_id = :series_id AND x = :x AND y = :y AND z = :z AND dev = :dev AND count < {$count_limit} LIMIT 1;");

                    $parameters_row = array(':series_id'  => $series_db->id,
                                            ':x'          => $series_db->x,
                                            ':y'          => $series_db->y,
                                            ':z'          => $series_db->z,
                                            ':dev'        => $series->dev); // TODO: check this (remove the dev from series)

                    $stmt_select_row->execute($parameters_row);
                    $series_db->row = $stmt_select_row->fetch(\PDO::FETCH_OBJ);

                    if ($first && $series_db->row==NULL) {
                        self::debug_X('There were no rows before.');
                        $stmt_insert_row = $conn->prepare("INSERT INTO {$rows_table} VALUES (DEFAULT, :x, :y, :z, :dev, DEFAULT, (SELECT counter(:domain)), :series_id, DEFAULT);");
                        $stmt_insert_row->execute(array_merge($parameters_row, array(':domain' => $this->_domain)));
                        $first = false;
                    } else {
                        break;
                    }

                } while (true);

                self::debug_X('Inserting StaticSmartDataBatch on Series ['.$series_db->id.'] row['.$series_db->row->id.']');

                $is_digital = $series->unit()->is_digital();

                $entry_type = \Cassandra\Type::userType(
                    'value',      ($is_digital) ? \Cassandra\Type::blob() : \Cassandra\Type::double(),
                    'error',      \Cassandra\Type::tinyint(),
                    'confidence', \Cassandra\Type::tinyint(),
                    'gw',         \Cassandra\Type::bigint()
                );

                // TODO: alter the cassandra table's name to Config file
                $cassandra_table  = $this->_domain.'_v1_1_';
                $cassandra_table .= $is_digital ? 'digital' : 'si';
                // -----

                // change this
                // TODO: get this values from the smartdata object
                $row_id     = new \Cassandra\Bigint($series_db->row->row_id);
                $error      = new \Cassandra\Tinyint(0);
                $confidence = new \Cassandra\Tinyint(0);
                $gw         = new \Cassandra\Bigint($this->_gw_id);
                // change this

                $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, $this->_username, $this->_password, Config::config()::CASSANDRA_KEYSPACE);
                $insert_stmt = $session->prepare("INSERT INTO {$cassandra_table} (id, timestamp, entry) VALUES (?, ?, ?);");
                $batch = new \Cassandra\BatchStatement(\Cassandra::BATCH_LOGGED);
                $count = 0;
                foreach ($smart_data as $key => $data) {
                    if ($series_db->workflow ?? false) {
                        $_data = clone $data;
                        $_data = $this->_workflow_input($_data, $series_db->workflow);
                    } else {
                        $_data = $data;
                    }
                    $batch->add($insert_stmt, array($row_id,
                                                    new \Cassandra\Varint($data->time),
                                                    $entry_type->create(
                                                        'value',      ($is_digital) ? new \Cassandra\Blob($_data->value) : (double) $_data->value,
                                                        'error',      $error,
                                                        'confidence', $confidence,
                                                        'gw',         $gw
                                                    )));
                    $count++;
                    if($count == 400){
                        $session->execute($batch);
                        $count = 0;
                        $batch = new \Cassandra\BatchStatement(\Cassandra::BATCH_LOGGED);
                    }
                }

                if($count > 0){
                    $session->execute($batch);
                    $count = 0;
                }

                $update_stmt = $conn->prepare("UPDATE {$rows_table} SET count = count + :smart_data_count WHERE id = :id LIMIT 1;");
                $update_stmt->execute(array('id' => $series_db->row->id, 'smart_data_count' => $smart_data_count));
            }
        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }

        return $sucess;
    }

    private function _insertStaticSmartDataBlob(Series $series_db, $msd) : bool {
        if ($series_db == NULL) {// || (($series_db->count + $smart_data_count) >= Config\CASSANDRA_MAX_ROW_SIZE) ) {
            throw new Exception\NoMatchingSeriesException('No matching series');
        }
        if($series_db->r != 0) {
            throw new Exception\NoMatchingSeriesException('Series must have r == 0'); // TODO: change this exception type
        }
        $sucess = true;
        try {
             // Execute input workflow on MultiSmartData
             if (($msd instanceof MultiSmartData) && ($series_db->workflow ?? false)){
                $_REQUEST['stats']['before_workflow'] =  intval(1000000 * microtime(true));
                $_msd = clone $msd;
                $msd = $this->_workflow_tth_input($_msd, $series_db->workflow);
             }
             // Series is received, select was already made before this method was called. 
             $bytes = 4;
             if ($msd->encoding == 'I64' || $msd->encoding == 'D64') {
                 $bytes = 8;
             }
             $_REQUEST['stats']['before_cassandra_insert'] =  intval(1000000 * microtime(true));
             $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, Config::config()::MYSQL_PASSWORD);
             $rows_table = '`rows`';
             $smart_data_count = strlen($msd->blob) / $bytes;
             $t_final = $msd->time + (( $smart_data_count - 1) * $msd->period);
             $count_limit = Config::config()::CASSANDRA_MAX_ROW_SIZE;
             $series_db->row = NULL;
             $first = true;
             do {
                 $stmt_select_row = $conn->prepare("SELECT * FROM {$rows_table} WHERE series_id = :series_id AND x = :x AND y = :y AND z = :z AND dev = :dev AND count < {$count_limit} LIMIT 1;");
                 $parameters_row = array(':series_id'  => $series_db->id,
                                         ':x'          => $series_db->x,
                                         ':y'          => $series_db->y,
                                         ':z'          => $series_db->z,
                                         ':dev'        => $msd->dev);

                 $stmt_select_row->execute($parameters_row);
                 $series_db->row = $stmt_select_row->fetch(\PDO::FETCH_OBJ);
                 if ($first && $series_db->row==NULL) {
                     self::debug_X('There were no rows before.');
                     $parameters_row = array_merge($parameters_row, array(':domain' => $this->_domain));
                     $stmt_insert_row = $conn->prepare("INSERT INTO {$rows_table} VALUES (DEFAULT, :x, :y, :z, :dev, DEFAULT, (SELECT counter(:domain)), :series_id, DEFAULT);");
                     $stmt_insert_row->execute($parameters_row);
                     $first = false;
                 } else {
                     break;
                 }
             } while (true);

            self::debug_X('Inserting StaticSmartDataBLOB on Series ['.$series_db->id.'] row['.$series_db->row->id.']');

            $entry_type = \Cassandra\Type::userType(
                 'period',       \Cassandra\Type::int(),
                 'gateway',      \Cassandra\Type::bigint(),
                 'uncertainty',  \Cassandra\Type::int(),
                 'data',         \Cassandra\Type::blob()
             );

             $cassandra_table  = $this->_domain.'_v1_1_si_blob';

             // change this
             // TODO: get this values from the smartdata object
             $row_id     = new \Cassandra\Bigint($series_db->row->row_id);
             $error      = new \Cassandra\Tinyint(0);
             $confidence = 0;
             $gw         = new \Cassandra\Bigint(-1);
             // change this

             $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, $this->_username, $this->_password, Config::config()::CASSANDRA_KEYSPACE);
             $insert_stmt = $session->prepare("INSERT INTO {$cassandra_table} (id, timestamp, entry) VALUES (?, ?, ?);");
             $batch = new \Cassandra\BatchStatement(\Cassandra::BATCH_LOGGED);
             $batch->add($insert_stmt, array($row_id,
                         new \Cassandra\Varint($msd->time),
                         $entry_type->create(
                                 'period',       $msd->period,
                                 'gateway',      $gw,
                                 'uncertainty',  $confidence,
                                 'data',         new \Cassandra\Blob($msd->blob)
             )));
             $session->execute($batch);
             $update_stmt = $conn->prepare("UPDATE {$rows_table} SET count = count + :smart_data_count WHERE id = :id LIMIT 1;");
             $update_stmt->execute(array('id' => $series_db->row->id, 'smart_data_count' => $smart_data_count));
             $_REQUEST['stats']['after_cassandra_insert'] =  intval(1000000 * microtime(true));
             $_REQUEST['stats']['domain'] = $this->_domain;
        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }

        return $sucess;
    }

 private function _insertTrackerSmartDataBlob(Tracker $trk, $msd) : bool {
        $sucess = true;
        try {
             if (($msd instanceof MultiSmartData) && ($trk->workflow ?? false)){
                $_msd = clone $msd;
                $msd = $this->_workflow_tth_input($_msd, $trk->workflow);
             }
             $conn = self::_mysqlConnect( Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, 
	                                 Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, 
                                         Config::config()::MYSQL_PASSWORD);
             $table = Config::config()::MySQL_Table($trk->version);
	     $bytes = 4;
             if (($msd->encoding ?? false) && ( $msd->encoding == 'I64' || $msd->encoding == 'D64')) {
                 $bytes = 8;
             }
             $rows_table = "smartdata_v1.trackers";
             $smart_data_count = strlen($msd->blob) / $bytes;
             $t_final = $msd->time + (( $smart_data_count - 1) * $msd->period);
             $count_limit = Config::config()::CASSANDRA_MAX_ROW_SIZE;
             if ($trk->row_id == 0) {
                  $create_row = $conn->prepare("SELECT counter(:domain)");
                  $params = array(':domain'  => $trk->domain);
                  $res = $create_row->execute($params);
                  $trk->row_id = $stmt_select_row->fetch(\PDO::FETCH_OBJ);
             }
            self::debug_X('Inserting MobileSmartDataBLOB on Series ['.$trk->id.'] row['.$trk->row_id.']');
            $entry_type = \Cassandra\Type::userType(
		 'x',            \Cassandra\Type::int(),
	         'y',            \Cassandra\Type::int(),
                 'z',            \Cassandra\Type::int(),
                 'period',       \Cassandra\Type::int(),
                 'gateway',      \Cassandra\Type::bigint(),
                 'uncertainty',  \Cassandra\Type::int(),
                 'data',         \Cassandra\Type::blob()
             );
             $cassandra_table  = $this->_domain.'_v1_2_si_blob';
             $row_id     = new \Cassandra\Bigint($trk->row_id);
             $confidence = $trk->uncertainty ?? 0;
             if ($confidence < 0) {
                 $confidence = 0;
             }
             $gw         = new \Cassandra\Bigint($msd->gw ?? -1);    
             $session = self::_cassandraConnect( Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT,
                                                 $this->_username, $this->_password, Config::config()::CASSANDRA_KEYSPACE );
             $insert_stmt = $session->prepare("INSERT INTO {$cassandra_table} (id, timestamp, entry) VALUES (?, ?, ?);");
             $batch = new \Cassandra\BatchStatement(\Cassandra::BATCH_LOGGED);
             $batch->add($insert_stmt, array($row_id,
                         new \Cassandra\Varint($msd->time),
                         $entry_type->create(
                                 'x',       $msd->x,
                                 'y',       $msd->y,
                                 'z',       $msd->z,
                                 'period',  $msd->period,
                                 'gateway', $gw,
                                 'uncertainty', $confidence,
                                 'data',         new \Cassandra\Blob($msd->blob)
             )));
             $session->execute($batch);
             $update_stmt = $conn->prepare("UPDATE {$rows_table} SET count = count + :smart_data_count WHERE id = :id LIMIT 1;");
             $update_stmt->execute(array('id' => $trk->id, 'smart_data_count' => $smart_data_count));
        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }
        return $sucess;
    }

    private function _insertMobileSmartDataBatch(Tracker $trkr, MobileSmartData ...$smart_data) : bool {

        $sucess = true;
        $smart_data_count = count($smart_data);

        try {
            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, Config::config()::MYSQL_PASSWORD);
            $table = Config::config()::MySQL_Table($trkr->version);
            $series_db = NULL;
	        $parameters = array( ':id' => $trkr->id,
                                 ':version' => $trkr->version,
                                 ':signature' => $trkr->signature ,
                                 ':sigsha'   => hash('sha256',$trkr->signature )
                                );
            $stmt_select = $conn->prepare("SELECT * FROM {$table} WHERE id = :id  AND version = :version AND signature in ( :signature , :sigsha ) LIMIT 1;");
            $stmt_select->execute($parameters);
            $series_db = $stmt_select->fetch(\PDO::FETCH_OBJ);

            if ($series_db == NULL){// || (($series_db->count + $smart_data_count) >= Config\CASSANDRA_MAX_ROW_SIZE) ) {
                throw new Exception\NoMatchingSeriesException('No matching series');
            }else{
                $count_limit = Config::config()::CASSANDRA_MAX_ROW_SIZE;

                self::debug_X('Inserting MobileSmartDataBatch on Trackers ['.$series_db->id.'] row['.$series_db->row_id.']');

                $is_digital = Unit::is_digital_unit($trkr->unit);

                $entry_type = \Cassandra\Type::userType(               
                    'value',      ($is_digital) ? \Cassandra\Type::blob() : \Cassandra\Type::double(),
                    'error',      \Cassandra\Type::tinyint(),
                    'confidence', \Cassandra\Type::tinyint(),
                    'x',          \Cassandra\Type::int(),
                    'y',          \Cassandra\Type::int(),
                    'z',          \Cassandra\Type::int(),     
                    'gw',         \Cassandra\Type::bigint()
                );

                // TODO: alter the cassandra table's name to Config file
                $cassandra_table  = $this->_domain.'_v1_2_';
                $cassandra_table .= $is_digital ? 'digital' : 'si';
                // -----

                // change this
                // TODO: get this values from the smartdata object
                $row_id     = new \Cassandra\Bigint($series_db->row_id);
                $error      = new \Cassandra\Tinyint(0);
                $confidence = new \Cassandra\Tinyint(0);
                $gw         = new \Cassandra\Bigint($this->_gw_id);
                // change this

                $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, $this->_username, $this->_password, Config::config()::CASSANDRA_KEYSPACE);
                $insert_stmt = $session->prepare("INSERT INTO {$cassandra_table} (id, timestamp, entry) VALUES (?, ?, ?);");
                $batch = new \Cassandra\BatchStatement(\Cassandra::BATCH_LOGGED);
	        $count = 0;
                foreach ($smart_data as $key => $data) {
                    if ($series_db->workflow ?? false) {
                        $_data = clone $data;
                        $_data = $this->_workflow_input($_data, $series_db->workflow);
                    } else {
                        $_data = $data;
                    }
                    $batch->add($insert_stmt, array($row_id,
                                                    new \Cassandra\Varint($data->time),
                                                    $entry_type->create(
                                                    'x',          $_data->x,
                                                    'y',          $_data->y,
                                                    'z',          $_data->z,
                                                    'value',      ($is_digital) ? new \Cassandra\Blob($_data->value) : (double) $_data->value,
                                                    'error',      $error,
                                                    'confidence', $confidence,
                                                    'gw',         $gw
                                                    )));
                    $count++;
                    if($count == 400){
                        $session->execute($batch);
                        $count = 0;
                        $batch = new \Cassandra\BatchStatement(\Cassandra::BATCH_LOGGED);
         	    }
                }

                if($count > 0){
                    $session->execute($batch);
                    $count = 0;
                }

                $update_stmt = $conn->prepare("UPDATE {$table} SET count = count + :smart_data_count WHERE id = :id LIMIT 1;");
                $update_stmt->execute(array('id' => $series_db->row_id, 'smart_data_count' => $smart_data_count));
            }
        } catch (\Exception $e) {
            Logger::exception($e);
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }

        return $sucess;
    }

    public function query(Series $series, $aggr=null, $options=null) {
        $response_json = array('series' => array());
        $sd_count = 0;
        try {

            $parameters = array(':version'  => $series->version,
                                ':t0'       => $series->t0,
                                ':t1'       => $series->t1,
                                ':domain'   => $this->_domain);

            $series_table = Config::config()::MySQL_Table($series->version);
            $rows_table = '`rows`'; // TODO: move to Config file (row table name)

            $query_series = "SELECT * FROM {$series_table} WHERE version = :version AND ( (:t0 >= t0 AND :t0 <= t1) OR (:t0 <= t0 AND :t1 >= t1) OR (:t1 >= t0 AND :t1 <= t1 ) )";

            if($series->unit != null){
                $parameters = array_merge($parameters, array(':unit' => $series->unit));
                $query_series .= ' AND unit = :unit';
            }

            $invalid_sphere = (int)($series->x === null) +
                              (int)($series->y === null) +
                              (int)($series->z === null) +
                              (int)($series->r === null);

            if ($invalid_sphere != 0 && $invalid_sphere != 4) {
                throw new BadRequestException("Invalid sphere parameters");
            } else if($invalid_sphere == 0) {
                $parameters = array_merge($parameters, array(':x'    => $series->x,
                                                             ':y'    => $series->y,
                                                             ':z'    => $series->z,
                                                             ':r'    => $series->r));

                $query_series .= ' AND SQRT(POW(x - :x, 2) + POW(y - :y, 2) + POW(z - :z, 2)) <= :r + r';
                $filter_sphere = true;
            } else {
                $filter_sphere = false;
            }
            //workflow removed from comparation as comment above
            $query_series .= ' AND domain = :domain;';
            
            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, Config::config()::MYSQL_PASSWORD);
            $stmt_select = $conn->prepare($query_series);
            $stmt_select->execute($parameters);
            $series_set = $stmt_select->fetchAll(\PDO::FETCH_OBJ);
            if ($series->unit == 0x00000001) {
                $mv_type = 8;
            } else {
                $mv_type = ($series->unit & 0x60000000) >> 28;
            }
            if($series_set) {
                $selected_series = array_column($series_set, null, 'id');
                // TODO: filter the xyz here
                $query_rows = "SELECT * FROM {$rows_table} WHERE series_id IN (".implode(',',array_column($series_set, 'id')).") AND dev = :dev;";
                $stmt_select = $conn->prepare($query_rows);
                $stmt_select->execute(array(':dev' => $series->dev));

                $rows_set = $stmt_select->fetchAll(\PDO::FETCH_OBJ);
                $selected_rows = array_column($rows_set, null, 'row_id');

                $id_set = \Cassandra\Type::collection(\Cassandra\Type::bigint())->create();
                foreach ($selected_rows as $row_id => $row_fields){
                    $id_set->add(new \Cassandra\Bigint($row_id));
                }

                $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, $this->_username, $this->_password, Config::config()::CASSANDRA_KEYSPACE);

                $is_digital  = $series->unit()->is_digital();
                if($is_digital) {
                    $type = $_GET['type'] ?? $options->type ?? (PHP_INT_MAX)>>29;
                    $type = $type<<29;
                }

                // TODO: move the cassandra table's name to Config file
                $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, $this->_username, $this->_password, Config::config()::CASSANDRA_KEYSPACE);             
                $cassandra_table  = $this->_domain.'_v1_1_';
                if (isset($series->type) and $series->type === 'TTH') {
                    $cassandra_table .= 'si_blob';
                    $cass_select_stmt = $session->prepare("SELECT * FROM {$cassandra_table} WHERE (id IN ?) AND (timestamp <= ?)");
                    $result_page = $session->execute($cass_select_stmt,
                        array('arguments' => array($id_set, new \Cassandra\Varint($series->t1))));
                } else {
                    $cassandra_table .= $is_digital ? 'digital' : 'si';
                    $cass_select_stmt = $session->prepare("SELECT * FROM {$cassandra_table} WHERE (id IN ?) AND (timestamp >= ?) AND (timestamp <= ?);");
                    $result_page = $session->execute($cass_select_stmt,
                        array('arguments' => array($id_set, new \Cassandra\Varint($series->t0), new \Cassandra\Varint($series->t1))));
                }
                $index = 0;
                $aggregator = ($this->internal()) ? null : Aggregator::new($aggr);
                
                $downSample = false;
                $step = 1;
                $point_number = 0;

                if ( ($aggregator != null && isset($aggregator->name) && $aggregator->name == 'downSample') ) {
                     $step = $aggregator->spacing; 
                     $downSample = true;
                } 
                
                while($result_page && $sd_count < Config::config()::POINTS_LIMIT) {
                    foreach ($result_page as $column) {
                        $row_id = (int)$column['id'];
                        $entry = $column['entry'];
                        $x = $selected_rows[$row_id]->x;
                        $y = $selected_rows[$row_id]->y;
                        $z = $selected_rows[$row_id]->z;
                        if($is_digital and ($series->type != 'TTH')){
                            $time  = (string)$column['timestamp']->value(); //bigint
                            $value_blob = $entry->get('value')->toBinaryString();
                            $value = null;
                            switch ($type) {
                                case Unit::I32:
                                    $value = Unpack::uInt32($value_blob);
                                    break;
                                case Unit::I64:
                                    $value = Unpack::uInt64($value_blob);
                                    break;
                                case Unit::F32:
                                    $value = Unpack::f32($value_blob);
                                    break;
                                case Unit::D64:
                                    $value = Unpack::d64($value_blob);
                                    break;
                                default:
                                    if(strlen($value_blob) == 1)
                                        $value = Unpack::uInt8($value_blob);
                                    break;
                            }
                            if($value === null){
                                if($value_blob != null)
                                    $value = $value_blob; // TODO: remove this in the new API
                                else
                                    $value = -400; // TODO: remover isso quando valores inválidos não puderem mais ser inseridos
                            }
                        } else if ($series->type === 'TTH') {                                
                            $time  = (string) $column['timestamp']->value(); //bigint
                            $bin = $entry->get('data')->toBinaryString();
                            $period = $entry->get('period');
                            $ts = (int) $time;
                            $per = (int) $period;
                            switch ($mv_type) {
                                case 0 : $pack_code = 'V*'; break;
                                case 2 : $pack_code = 'P*'; break;
                                case 4 : $pack_code = 'g*'; break;
                                case 6 : $pack_code = 'e*'; break;
                                case 8 : $pack_code = 'C*'; break; // digital switches
                            }
                            $values_array = unpack($pack_code, $bin);
                            $smartdata = array();
                            $pos = 1;
                            $gw    = (int)$entry->get('gateway'); 
                            $conf  = (int)$entry->get('uncertainty');
                            $error = 0;
                            $d = $series->dev;                   
                            foreach ($values_array as $value) {
                                if ((!$downSample || ($point_number % $step) == 0) && 
                                   ($ts >= $series->t0 && $ts <= $series->t1)) {
                                    $sd = new StaticSmartData($series->unit,$value,$error,$conf,$x,$y,$z,$ts,$d);
                                    $smartdata[$pos++] = $sd;
                                    $sd_count += 1;
                                } 
                                $point_number += 1;
                                $ts = $ts + $per;
                                if ($sd_count > Config::config()::POINTS_LIMIT) {
                                    break;
                                }
                            }
                        } else {
                            if (!$downSample || ($point_number % $step) == 0) { 
                                $time  = (string)$column['timestamp']->value(); //bigint
                                if ($is_digital) {
                                    $value = base64_encode( $entry->get('value')->toBinaryString()); // its a digital value, possibly a blob!
                                } else {
                                    $value = (double)$entry->get('value'); // double
                                }
                                $point_number += 1;
                            } else {
                                $point_number += 1;
                                continue;
                            }
                        }
                        // TODO: check this filter here
                        if ( $filter_sphere && !$series->contains_point($x, $y, $z) ) {
                            continue;
                        }
                        // TODO: what about digital entries?
                        if ($series->type !== 'TTH') {
                            $error = (int)$entry->get('error');             //tinyint
                            $conf  = (int)$entry->get('confidence');        //tinyint
                            $d     = $series->dev;  
                            $gw    = (int)$entry->get('gw');                //int (TO DO GET GATEWAY)
                            switch ($series->version) {
                                case SmartData::STATIC_VERSION:
                                    $smartdata = new StaticSmartData($series->unit,$value,$error,$conf,$x,$y,$z,$time,$d,$gw);
                                    break;
                                case SmartData::MOBILE_VERSION:
                                    $smartdata = new MobileSmartData($series->unit,$value,$error,$conf,$x,$y,$z,$time,$d,$gw, null);
                                    break;
                                default:
                                    continue 2;
                                    break;
                            }
                        }
                        if($aggregator != null && !$downSample) {
                            $smartdata = $aggregator->aggregate($smartdata);
                        }

                        if($smartdata != null) {
                            if(is_array($smartdata)){
                                if ($series->type === 'TTH') {
                                    $md = array();
                                    $md["version"] = "1.1";
                                    $md["unit"] = $series->unit;
                                    $md["x"] = $series->x;
                                    $md["y"] = $series->y;
                                    $md["z"] = $series->z;
                                    $md["r"] = 0;
                                    $md["type"] = "TTH";
                                    $md["period"] = $period;
                                    $md["uncertainty"] = 0;
                                    $md["dev"] = $series->dev;
                                    $md["t0"] = array_values($smartdata)[0]->time;
                                    $md["gateway"] = -1;
                                    $md["datapoints"] = array();
                                    $idx = 0;
                                    foreach ($smartdata as $key => $data) {
                                        $md["datapoints"][$idx]["value"] = $data->value;
                                        $idx += 1;
                                    }
                                    $response_json['series'][$index++]['MultiValueSmartData'] = $md;
                                } else {
                                    foreach ($smartdata as $key => $data) {
                                        if ((!$downSample) || ($downSample && (($point_number % $step) == 0)))  {
                                            $response_json['series'][$index++] = $data->toArray();
                                            if ($index >= Config::config()::POINTS_LIMIT) break;
                                        }
                                        $point_number += 1;
                                    }
                                }
                            } else {
                                $response_json['series'][$index++] = $smartdata->toArray();
                                if($this->internal()){
                                    $response_json['series'][$index-1]['slice'] = $row_id;
                                    $response_json['series'][$index-1]['gw'] = $gw;
                                }
                            }
                        } else {
                            continue;
                        }
                        if ($index >= Config::config()::POINTS_LIMIT) break 2;
                    }
                    $result_page = $result_page->nextPage();
                    if(!$result_page && isset($aggregator) && $index < Config::config()::POINTS_LIMIT){
                        $smartdata = $aggregator->finish();
                        if($smartdata != null) {
                            $response_json['series'][$index++] = $smartdata->toArray();
                        }
                    }
                }
                $session->close();
                usort($response_json['series'], function ($a, $b) {
                    $a_timestamp  = gmp_init($a['timestamp']);
                    $b_timestamp  = gmp_init($b['timestamp']);
                    return gmp_cmp($a_timestamp, $b_timestamp);
                 });
            } 
        } catch (\Exception $e) {
            Logger::exception($e);
            $response_json = NULL;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }
        $workflow = $series->workflow ?? 0;
        if($response_json !== NULL && $workflow !== 0) {
            self::debug_X('Enter output workflow');
            $_response_json = $response_json;
            $response_json = $this->_workflow_output($_response_json, $workflow, $series);
        }
        return $response_json;
    }

    public function finish(Series $series, $parameter=null) {
        try {
            $sucess = true;
            $parameters = array(':version'  => $series->version,
                                ':unit'     => $series->unit,
                                ':t1'       => 9999999999999999, // Series must be 'open'
                                ':type'     => $series->type ?? 'OLD',
                                ':domain'   => $this->_domain);
            $series_table = Config::config()::MySQL_Table($series->version);
            $query_series = "SELECT * FROM {$series_table} WHERE version = :version AND unit = :unit ";
            if ($series->version === SmartData::STATIC_VERSION) { 
                $query_series = $query_series . " AND x = :x AND y = :y AND z = :z ";
                $parameters = array_merge($parameters, array(':x'=> $series->x));
                $parameters = array_merge($parameters, array(':y'=> $series->y));
                $parameters = array_merge($parameters, array(':z'=> $series->z));
            } 
            $query_series = $query_series . " AND t1 = :t1 AND type = :type AND domain = :domain ";
            if ($series->period > 0) {
                 $query_series = $query_series . " AND period = :period";
                 $parameters = array_merge($parameters, array(':period'=> $series->period));
            }
            if ($series->version === SmartData::MOBILE_VERSION) {
                 if (!property_exists($series,"signature") || $series->signature == 0)
                      throw new Exception\InternalException('Signature must be provided in Mobile SmartData.');
                 $query_series .= " AND signature in ( :signature, :sigsha ) AND dev = :dev";
                 $mobile_params = array(
                    ':signature'=> $series->signature,
                    ':sigsha' => hash('sha256',$series->signature),
                    ':dev' => $series->dev );
                 $parameters = array_merge($parameters, $mobile_params );
            }
            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME, Config::config()::MYSQL_USERNAME, Config::config()::MYSQL_PASSWORD);
            $stmt_select = $conn->prepare($query_series);
            $stmt_select->execute($parameters);
            $aux = $stmt_select->fetchAll( \PDO::FETCH_OBJ );
            $cont = 0;
            foreach ( $aux as $x ) {
                $upd_query = "UPDATE {$series_table} SET t1 = :t1 ";
                $pars = array(':t1' => $series->t1, ':id' => $x->id);
                if (isset($series->event)){
                    $pars = array_merge($pars, array(":event" =>  $series->event));
                    $upd_query = $upd_query . ", event = :event ";
                }
                $upd_query = $upd_query . " where id = :id";
                $update_stmt = $conn->prepare($upd_query);
                $update_stmt->execute($pars);
                $cont += 1;
            }
            if ($cont === 0) {
                throw new Exception\InternalException('Series not found or already closed.');
            }
        } catch(Exception $ex) {
            $sucess = false;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }
        return $sucess;
    }    

    public function search(Series $series, $parameter=null) {
        // Assuming id is passed as a Workflow and no daemon is necessary
        $response_json = array('series' => array());
        $id = $series->workflow ?? 0;

        try {
            $_root  = __DIR__;
            $_f = "search{$id}";
            $pushd = "pushd {$_root}/../workflow/{$this->_domain} > /dev/null";
            $popd  = 'popd > /dev/null';
            $json_parameter = json_encode($parameter);
            $json_series = json_encode($series->toArray($series));
            $return = shell_exec("{$pushd}; ./{$_f} '{$json_series}' '{$json_parameter}'; {$popd};");
            $json = json_decode($return, true, 512, JSON_BIGINT_AS_STRING);
            if (json_last_error() !== JSON_ERROR_NONE) 
                throw new Exception\InternalException('Error while processing filter!');
            $response_json = $json;
        } catch (\Exception $e) {
            Logger::exception($e);
            $response_json = NULL;
        } // finally {}
        if ( ! isset( $series->signature ) )
            $response_json = Anonymizer::anonymizeSeries($response_json); //TODO mudar para referencia posteriormente
        return $response_json;
    }

    public function track(Series $series, $aggr=null) {
        $sd_count = 0;
        $response_json = NULL;
        try {
            if ($series->t0 < 0 || $series->t1 < 0 || $series->t0 >= $series->t1 ) {
                throw new BadRequestException("Invalid Time Interval.");
            }
            $parameters = array(':version'   => $series->version,
                                ':unit'      => $series->unit,
                                ':signature' => $series->signature,
                                ':sigsha'    => hash('sha256',$series->signature),
                                ':period'    => $series->period,
                                ':dev'       => $series->dev,
                                ':type'      => $series->type ?? 'OLD',
                                ':t0'        => $series->t0,
                                ':t1'        => $series->t1,
                                ':domain'    => $this->_domain);

            $table = Config::config()::MySQL_Table($series->version);

            $query = "SELECT * FROM {$table}
                      WHERE version = :version 
                      AND unit = :unit 
                      AND dev = :dev 
                      AND type = :type 
                      AND t0 <= :t1 
                      AND t1 >= :t0 
                      AND domain = :domain";
	        if( isset( $series->signature ) && $series->signature != null ){
                $query .= ' AND signature in ( :signature , :sigsha ) ';
	        }else{
		        unset($parameters[':signature']);
		        unset($parameters[':sigsha']);
            }

            if( $series->period != null && $series->period > 0){
                $query .= ' AND period = :period ';
	        }else{
		        unset($parameters[':period']);
            }
            $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME,
                                        Config::config()::MYSQL_PORT,
                                        Config::config()::MYSQL_DBNAME,
                                        Config::config()::MYSQL_USERNAME,
                                        Config::config()::MYSQL_PASSWORD);
	        $select_stmt = $conn->prepare($query.';');
            $select_stmt->execute($parameters);
            $aux = $select_stmt->fetchAll();
            $id_set = \Cassandra\Type::collection(\Cassandra\Type::bigint())->create();
            //$found = false;

            $row2signature = [];
            foreach ( $aux as $x ){
                $id_set->add(new \Cassandra\Bigint($x['row_id']));
                $row2signature[$x['row_id']] = $x['signature'];
            }

            $is_digital = $series->unit()->is_digital();
            $session = self::_cassandraConnect(Config::config()::CASSANDRA_SERVERNAME, Config::config()::CASSANDRA_PORT, $this->_username, $this->_password, Config::config()::CASSANDRA_KEYSPACE);

            $cassandra_table = "{$this->_domain}";
            $cassandra_table .= ($series->version == 17) ? "_v1_1" : "_v1_2";
    	    if ($series->type == 'TTH') {
                $cassandra_table .= "_si_blob";
            } else {
                $cassandra_table .= ($is_digital) ? "_digital" : "_si";
            }
            if ($series->type == 'TTH') {
                $cass_select_stmt = $session->prepare("SELECT * FROM {$cassandra_table} WHERE (id IN ?) AND (timestamp <= ?) ");
                $cass_args = array('arguments' => array($id_set, new \Cassandra\Varint($series->t1)));
                $result_page = $session->execute($cass_select_stmt, $cass_args);
            } else {
                $cass_select_stmt = $session->prepare("SELECT * FROM {$cassandra_table} WHERE (id IN ?) AND (timestamp >= ?) AND (timestamp <= ?) ");
                $cass_args = array('arguments' => array($id_set, new \Cassandra\Varint($series->t0), new \Cassandra\Varint($series->t1)));
                $result_page = $session->execute($cass_select_stmt, $cass_args);
            }

            $index = 0;
            $response_json = array('series' => array());
            $aggregator = ($this->internal()) ? null : Aggregator::new($aggr);

            $downSample = false;
            $step = 1;
            $point_number = 0;
            if ( ($aggregator != null && isset($aggregator->name) && $aggregator->name == 'downSample') ) {
                 $step = $aggregator->spacing; 
                 $downSample = true;
            } 
            $mv_type = ($series->unit == 0x00000001) ? 8 : ($series->unit & 0x60000000) >> 28;

            Logger::debug('Type: '.$mv_type.' isDigital: '.$is_digital.' unit: '.$series->unit.' dev: '.$series->dev.' sig: '.$series->signature);
            $queries_signature = $series->signature;
            
            while($result_page && $sd_count < Config::config()::POINTS_LIMIT) {
                foreach ($result_page as $column) {
                    //TODO temos que mapear o row_id com o signature
                    $row_id = (int)$column['id'];
                    $return_signature = ($queries_signature == "0") ? $row2signature[$row_id] : $queries_signature ;
                    $entry = $column['entry'];
                    $x = $entry->get('x');
                    $y = $entry->get('y');
                    $z = $entry->get('z');
		            $time  = (string)$column['timestamp']->value(); //bigint
                    if ( ! $series->contains_point($x, $y, $z)) {
                        continue;
                    }
                    if($is_digital and ($series->type != 'TTH')){
                        $value_blob = $entry->get('value')->toBinaryString();
                        switch ($series->type) {
                            case Unit::I32:
                                $value = Unpack::uInt32($value_blob);
                                break;
                            case Unit::I64:
                                $value = Unpack::uInt64($value_blob);
                                break;
                            case Unit::F32:
                                $value = Unpack::f32($value_blob);
                                break;
                            case Unit::D64:
                                $value = Unpack::d64($value_blob);
                                break;
                            default:
                                if(strlen($value_blob) == 1)
                                    $value = Unpack::uInt8($value_blob);
                                break;
                        }
                        $value = base64_encode( $value_blob );
                        if($value === null){
                            // TODO: remove this in the new API (Scheffel)
                            // TODO: remover isso quando valores inválidos não puderem mais ser inseridos
                            $value = ( is_null($value_blob)) ? -400 : $value_blob ;
                        }
                    } else if ($series->type === 'TTH') {
                        $time  = (string) $column['timestamp']->value(); //bigint
                        $bin = $entry->get('data')->toBinaryString();
                        $period = $entry->get('period');
                        $ts = (int) $time;
                        $per = (int) $period;
                        switch ($mv_type) {
                            case 0 : $pack_code = 'V*'; break;
                            case 2 : $pack_code = 'P*'; break;
                            case 4 : $pack_code = 'g*'; break;
                            case 6 : $pack_code = 'e*'; break;
                            case 8 : $pack_code = 'C*'; break; // digital switches
                        }
                        $values_array = unpack($pack_code, $bin);
                        $smartdata = array();
                        $pos = 1;
                        $gw    = (int)$entry->get('gateway'); 
                        $conf  = $entry->get('uncertainty') ?? 0;
                        $error = 0;
                        $d = $series->dev;
                        foreach ($values_array as $value) {
                            if ((!$downSample || ($point_number % $step) == 0) && 
                               ($ts >= $series->t0 && $ts <= $series->t1)) {
                                $sd = new MobileSmartData($series->unit,$value,$error,$conf,$x,$y,$z,$ts,$d,$series->signature,$gw);
                                $smartdata[$pos++] = $sd;
                                $sd_count++;
                            }
                            $point_number += 1;
                            $ts = $ts + $per;
                            if ($sd_count > Config::config()::POINTS_LIMIT) {
                                break;
                            }
                        }
		            } else {
                        if (!$downSample || ($point_number % $step) == 0) {
                            $time  = (string)$column['timestamp']->value(); //bigint
                            if ($is_digital) {
                                $value = base64_encode( $entry->get('value')->toBinaryString()); // its a digital value, possibly a blob!
                            } else {
                                $value = (double)$entry->get('value'); // double
                            }
                            $point_number += 1;
			            } else {
                            $point_number += 1;
                            continue;
                        }
                    }

                    // TODO: what about digital entries?
                    if ($series->type !== 'TTH') {
                        $error = (int)$entry->get('error');             //tinyint
                        $conf  = (int)$entry->get('confidence');        //tinyint
                        $d     = $series->dev;
                        $gw    = (int)$entry->get('gw');                //int (TO DO GET GATEWAY)
                        switch ($series->version) {
                            case SmartData::STATIC_VERSION:
                                $smartdata = new StaticSmartData($series->unit,$value,$error,$conf,$x,$y,$z,$time,$d,$gw);
                                break;
                            case SmartData::MOBILE_VERSION:
                                
                                $smartdata = new MobileSmartData($series->unit, $value, $error, $conf, $x, $y, $z, $time, $d, $gw, $return_signature );
                                break;
                            default:
                                continue 2;
                        }
                    }
                    if($aggregator != null && !$downSample) {
                        $smartdata = $aggregator->aggregate($smartdata);
                    }
                    if($smartdata != null) {
                        if(is_array($smartdata)){
                            if ($series->type === 'TTH') {
                                $md = array();
                                $md["version"] = "1.2";
                                $md["unit"] = $series->unit;
                                $md["x"] = $series->x;
                                $md["y"] = $series->y;
                                $md["z"] = $series->z;
                                $md["r"] = 0;
                                $md["signature"] = $return_signature;
                                $md["t0"] = array_values($smartdata)[0]->time;
                                $md["type"] = "TTH";
                                $md["period"] = $period;
                                $md["uncertainty"] = 0;
                                $md["dev"] = $series->dev;
                                $md["gateway"] = -1;
                                $md["datapoints"] = array();
                                $idx = 0;
                                foreach ($smartdata as $key => $data) {
                                    $md["datapoints"][$idx]["value"] = $data->value;
                                    $idx += 1;
                                }
                                $response_json['series'][$index++]['MultiValueSmartData'] = $md;
                            } else {
                                foreach ($smartdata as $key => $data) {
                                    if ((!$downSample) || ($downSample && (($point_number % $step) == 0)))  {
                                        $response_json['series'][$index++] = $data->toArray();
                                        if ($index >= Config::config()::POINTS_LIMIT) break;
                                    }
                                }
                            }
                        } else {
                            $response_json['series'][$index++] = $smartdata->toArray();
                            if($this->internal()){
                                $response_json['series'][$index-1]['slice'] = $row_id;
                                $response_json['series'][$index-1]['gw'] = $gw;
                            }
                            $point_number += 1;
                        }
                    } else {
                        continue;
                    }
                    if ($index >= Config::config()::POINTS_LIMIT) break 2;
                }
                $result_page = $result_page->nextPage();
                if(!$result_page && isset($aggregator) && $index < Config::config()::POINTS_LIMIT){
                    $smartdata = $aggregator->finish();
                    if($smartdata != null) 
                        $response_json['series'][$index++] = $smartdata->toArray();
                }
            }
            $session->close();
        } catch (\Exception $e) {
            Logger::exception($e);
            $response_json = NULL;
        } finally {
            if (isset($session))
                $session->close();
            $conn = null;
        }
        $workflow = $series->workflow ?? 0;
        if($response_json !== NULL && $workflow !== 0) {
            self::debug_X('Enter output workflow');
            $_response_json = $response_json;
            $response_json = $this->_workflow_output($_response_json, $workflow, $series);
        }
        if ( $queries_signature == "0" )
            $response_json = Anonymizer::anonymizeSeries($response_json); //TODO mudar para referencia posteriormente
        
        return $response_json;
    }
    
     public function track_multiple(MobileSmartData ...$msd) { 
        $sucess = true;
        $smart_data_count = count($msd);
        if ($smart_data_count < 1) 
            return false;

        // Multiple devices can be sent, grouping by device
        $dev_map = array();
        foreach ($msd as $sd) {
            $vals =  (array_key_exists( $sd->dev, $dev_map )) ? $dev_map[ $sd->dev ] : array();
            $vals[] = $sd;
            $dev_map[$sd->dev] = $vals;
        }

        // First: define max and min timestamp, and verify if a there is a series that contains all
        foreach ($dev_map as $smart_data) {
            $sd =  $smart_data[0];
            $min = $sd->time;
            $max = $min;
            foreach ($smart_data as $sd) {
                if ($sd->time > $max) {
                    $max = $sd->time;
                } else if ($sd->time < $min) {
                    $min = $sd->time; 
                }
            }
            try {
                //TODO Conexao dentro de um foreach???
               $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME,
                                           Config::config()::MYSQL_PORT,
                                           Config::config()::MYSQL_DBNAME,
                                           Config::config()::MYSQL_USERNAME,
                                           Config::config()::MYSQL_PASSWORD);
    
               $table = Config::config()::MySQL_Table($sd->version);
               $parameters = array(':version'    => $sd->version,
                                    ':unit'      => $sd->unit,
                                    ':period'    => $sd->period,
                                    ':signature' => $sd->signature,
                                    ':sigsha'    => hash('sha256',$sd->signature),
                                    ':type'      => $sd->type ?? 'OLD',
                                    ':t'         => $sd->time,
                                    ':dev'       => $sd->dev,
                                    ':domain'    => $this->_domain);
    
                $table = Config::config()::MySQL_Table($sd->version);
    
                $query = "SELECT * FROM {$table}
                          WHERE version = :version 
                          AND unit = :unit 
                          AND dev = :dev 
                          AND type = :type 
                          AND t0 <= :t 
                          AND t1 >= :t 
                          AND signature in ( :signature , :sigsha )
                          AND domain = :domain ";
                
                if( isset($sd->period) && $sd->period > 0){
                    $query .= ' AND period = :period ';
                }else{
                    unset($parameters[':period']);
                }
                    
                $stmt_select = $conn->prepare( $query . ';');
                $stmt_select->execute($parameters);
                $sdb = $stmt_select->fetch(\PDO::FETCH_OBJ);
                if ( is_null($sdb) || ! isset( $sdb->version ) ) 
                    throw new BadRequestException("Series was not found, check sent data.");
                $trkr = new Tracker( $sdb->version, $sdb->unit, $sdb->signature, $sdb->t0, $sdb->t1, $sdb->domain, $sdb->dev, 
                        $sdb->workflow, $sdb->type, $sdb->period, $sdb->uncertainty, $sdb->id, $sdb->row_id, $sdb->event);
                $sucess = $this->_insertMobileSmartDataBatch($trkr, ...$smart_data);

           } catch (\Exception $e) {
                Logger::exception($e);
                $sucess = false;
            } finally {
                if (isset($session))
                    $session->close();
                $conn = null;
            }
        }
        return $sucess;
    }
    
    public function describe($values) {
        foreach( $values as $descr) {
            $sucess = true;
            $getParams =  array(':version'    => $descr->version,
                                ':unit'       => $descr->unit,
                                ':x'          => $descr->x,
                                ':y'          => $descr->y,
                                ':z'          => $descr->z,
                                ':dev'        => $descr->dev,
                                ':domain'     => $this->_domain,
                                ':type'       => $descr->type,
                                ':period'     => $descr->period,
                                ':signature'  => $descr->signature,
                                ':sigsha'     => hash('sha256',$descr->signature)
                            );                              
            $parameters = array(':version'    => $descr->version,
                                ':unit'       => $descr->unit,
                                ':x'          => $descr->x,
                                ':y'          => $descr->y,
                                ':z'          => $descr->z,
                                ':dev'        => $descr->dev,
                                ':domain'     => $this->_domain,
                                ':type'       => $descr->type,
                                ':period'     => $descr->period,
                                ':signature'  => hash('sha256',$descr->signature),
                                ':description'=> $descr->description
                            );
            try {            
                $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME,
                 Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME,
                 Config::config()::MYSQL_USERNAME, 
                 Config::config()::MYSQL_PASSWORD);
                $stmt_select = $conn->prepare(
                    "SELECT id FROM dictionary  
                    WHERE version = :version 
                    AND unit = :unit 
                    AND x = :x 
                    AND y = :y 
                    AND z = :z 
                    AND dev = :dev 
                    AND domain = :domain 
                    AND type = :type 
                    AND period = :period 
                    AND signature in ( :signature , :sigsha );");
                $stmt_select->execute($getParams);
                $res = $stmt_select->fetch(\PDO::FETCH_OBJ);       
                if ($res != null && property_exists( $res, "id" ) ) {
                    $upd_params = array( "description" => $descr->description, "id" => $res->id,);
                    $stm_upd = $conn->prepare("UPDATE dictionary set description = :description WHERE id = :id");
                    $x = $stm_upd->execute($upd_params);
                } else { // Its a new description, insert            
                   $stmt_insert = $conn->prepare("
                   INSERT INTO dictionary( version, unit, x, y, z, dev, domain, type, period, signature, description) 
                   VALUES (:version, :unit, :x, :y, :z, :dev, :domain, :type, :period, :signature, :description);" );
                   $stmt_insert->execute($parameters);
                }
            } catch (\Exception $e) {
                Logger::exception($e);
                $sucess = false;
            } finally {
                $conn = null;
            }
        }
        return $sucess;
    }
    
     public function list_devices($series) {
          $result = array();
          $params = array( ":version" => $series->version, 
                           ":domain" => $this->_domain,
                           ":r" => $series->r,
                           ":x" => $series->x,
                           ":y" => $series->y,
                           ":z" => $series->z);
          $query = "SELECT * FROM dictionary WHERE 
                    version = :version  
                    AND domain = :domain 
                    AND :r >= (SQRT(POW(x - :x, 2) +  POW(y - :y, 2) + POW(z - :z, 2)))";
          if ($series->unit != null) {
             $query .= " and unit = :unit";
             $params[":unit"] = $series->unit;
          }
          if ($series->signature != null) {
             $query .= " and signature in ( :signature, :sigsha ) ";
             $params[":signature"] = $series->signature;
             $params[":sigsha"]    = hash('sha256',$series->signature );    
          }
          if ($series->type != null) {
             $query .= " and type = :type";
             $params[":type"] = $series->type;
          }
          if ($series->period != null && $series->period > 0) {
             $query .= " and period = :period";
             $params[":period"] = $series->period;
          } 
          try {            
                $conn = self::_mysqlConnect(Config::config()::MYSQL_SEVERNAME, 
                Config::config()::MYSQL_PORT, Config::config()::MYSQL_DBNAME,
                 Config::config()::MYSQL_USERNAME, 
                 Config::config()::MYSQL_PASSWORD);
                $stmt_select = $conn->prepare($query);
                $stmt_select->execute($params);
                $res = $stmt_select->fetchAll(\PDO::FETCH_OBJ);       
                foreach($res as $desc) {
                    $x =  array("version" => $desc->version,
                                "unit" => $desc->unit,
                                "x" => $desc->x,
                                "y" => $desc->y,
                                "z" => $desc->z,
                                "dev" => $desc->dev,
                                "type" => $desc->type,
                                "period" => $desc->period,
                                "signature" => $desc->signature,
                                "description" => $desc->description);
                    array_push($result, $x);
                }
            } catch (\Exception $e) {
                Logger::debug($e);
                $result = null;
            } finally {
                $conn = null;
            }
        return $result;
    }
}

switch (Config_Common::api_version()) {
    case Config_Common::VERSION_1_0: class Backend extends Backend_V1_0 {} break;
    case Config_Common::VERSION_1_1: class Backend extends Backend_V1_1 {} break;
    default:
        die ("Bad API version!\n");
}
