<?php

declare(strict_types=1);

require_once 'client.php';

use SmartDataContextAPI\SmartDataContextAPIClient;

const CLIENT_CERTIFICATE = 'client.pem';
const CLIENT_CERTIFICATE_KEY = 'client.key';
const API_URL = 'https://localhost/';

function generateRandomFile(string $filename, int $sizeInMb): void
{
    $sizeInBytes = $sizeInMb * 1024 * 1024;
    $data = random_bytes($sizeInBytes);
    file_put_contents($filename, $data);
}

# Create a new SmartDataContextClient, passing a custom URL for the API and the certificates to access the domain
# The parameter verifyCertificate allows using self-signed certificates for development and test environments.
# Should not be used in production ideally
$client = new SmartDataContextAPIClient(
    certFile: CLIENT_CERTIFICATE,
    keyFile: CLIENT_CERTIFICATE_KEY,
    url: API_URL,
    verifyCertificate: false
);

# Creates a new perene SmartDataContext (no start not end time). The minimum required parameters are the content
# and the features, with at least one feature (tags), which indicate tags to be associated with this SmartDataContext
# The function returns the create SmartDataContext id or throws an exception in case of error
$pereneSmartDataContextId = $client->createSmartDataContext(
    content: ['meta' => true],
    features: ['tags' => ['sampleTag', 'simulation']]
);
echo "Created a new SmartDataContext with id {$pereneSmartDataContextId}\n";

# Now that we have a smartDataContext we can associate it to a SmartDataUnit or to SmartDataSource
# A smartDataUnit is represented by its value
# A smartDataSource is represented by its sphere (stationary) or its signature (mobile)
# Either smartDataUnits or smartDataSources or both can be associated to a smartDataContext
# A list of smartDataContextIds could be passed as well
# The function returns the list of updated SmartDataContext object or the single SmartDataContext
$smartDataUnits = [0x84963924];
$smartDataSources = [[1, 1, 2, 10], 'aeccd287'];
$pereneSmartDataContext = $client->associateSmartDataContext(
    smartDataUnits: $smartDataUnits,
    smartDataSources: $smartDataSources,
    smartDataContextIds: $pereneSmartDataContextId
);
echo "Updated SmartDataContext: \n" . json_encode($pereneSmartDataContext, JSON_PRETTY_PRINT) . "\n";

# To unassociate just call the unassociateSmartDataContext following the same syntax
$smartDataSources = [[1, 1, 2, 10]];
$pereneSmartDataContext = $client->unassociateSmartDataContext(
    smartDataSources: $smartDataSources,
    smartDataContextIds: $pereneSmartDataContextId
);
echo "Updated SmartDataContext: \n" . json_encode($pereneSmartDataContext, JSON_PRETTY_PRINT) . "\n";

# To retrieve a SmartDataContext by its id
$pereneSmartDataContext = $client->getSmartDataContext(smartDataContextId: $pereneSmartDataContextId);
echo "Retrieved SmartDataContext: \n" . json_encode($pereneSmartDataContext, JSON_PRETTY_PRINT) . "\n";

# To find all SmartDataContext associated to a set of smartDataSources or smartDataUnits
$smartDataContextForSources = $client->findSmartDataContext(smartDataSources: $smartDataSources);
echo "Retrieved SmartDataContext for sources: " . count($smartDataContextForSources) . "\n" . json_encode($smartDataContextForSources, JSON_PRETTY_PRINT) . "\n";

$smartDataContextForUnits = $client->findSmartDataContext(smartDataUnits: $smartDataUnits);
echo "Retrieved SmartDataContext for units: " . count($smartDataContextForUnits) . "\n" . json_encode($smartDataContextForUnits, JSON_PRETTY_PRINT) . "\n";

$smartDataContextForUnitsOrSources = $client->findSmartDataContext(smartDataUnits: $smartDataUnits, smartDataSources: $smartDataSources);
echo "Retrieved SmartDataContext for sources and units: " . count($smartDataContextForUnitsOrSources) . "\n" . json_encode($smartDataContextForUnitsOrSources, JSON_PRETTY_PRINT) . "\n";

# To find all SmartDataContext associated to a set of smartDataSources or smartDataUnits under a time range
$temporalSmartDataContextId = $client->createSmartDataContext(
    content: ['meta' => true],
    features: ['tags' => ['sampleTag', 'simulation']],
    smartDataUnits: $smartDataUnits,
    t0: 15,
    t1: 25
);
$temporalSmartDataContext = $client->getSmartDataContext(smartDataContextId: $temporalSmartDataContextId);
echo "Retrieved temporal SmartDataContext: \n" . json_encode($temporalSmartDataContext, JSON_PRETTY_PRINT) . "\n";

$smartDataContextForUnitsOrSources = $client->findSmartDataContext(
    smartDataUnits: $smartDataUnits,
    smartDataSources: $smartDataSources,
    t0: 16,
    t1: 20
);
echo "Retrieved SmartDataContext between 16 and 20: " . count($smartDataContextForUnitsOrSources) . "\n" . json_encode($smartDataContextForUnitsOrSources, JSON_PRETTY_PRINT) . "\n";

# To find all SmartDataContext associated to a set of smartDataSources or smartDataUnits under a time range
$smartDataContextForUnitsOrSources = $client->findSmartDataContext(
    smartDataUnits: $smartDataUnits,
    smartDataSources: $smartDataSources,
    t0: 5,
    t1: 9
);
echo "Retrieved SmartDataContext between 5 and 9: " . count($smartDataContextForUnitsOrSources) . "\n" . json_encode($smartDataContextForUnitsOrSources, JSON_PRETTY_PRINT) . "\n";

# To do a generic query using the MongoDB syntax
$query = ['t0' => 15];
$queryResult = $client->querySmartDataContext(query: $query);
echo "Query result: \n" . json_encode($queryResult, JSON_PRETTY_PRINT) . "\n";

# To change the contents of a SmartDataContext
$temporalSmartDataContext = $client->updateSmartDataContext($temporalSmartDataContextId, content: ['meta' => false]);
echo "Updated temporal SmartDataContext: \n" . json_encode($temporalSmartDataContext, JSON_PRETTY_PRINT) . "\n";

# To add an unstructured content to a SmartDataContext
generateRandomFile('data.bin', 3);
$storageObjectId = $client->addUnstructuredDataFromFile($temporalSmartDataContextId, 'data.bin');
echo "Stored object id is {$storageObjectId}\n";
$temporalSmartDataContext = $client->getSmartDataContext($temporalSmartDataContextId);
echo "Updated temporal SmartDataContext: \n" . json_encode($temporalSmartDataContext, JSON_PRETTY_PRINT) . "\n";

generateRandomFile('data2.bin', 3);
$storageObjectId2 = $client->addUnstructuredDataFromFile($temporalSmartDataContextId, 'data2.bin');
echo "Stored object id is {$storageObjectId2}\n";
$temporalSmartDataContext = $client->getSmartDataContext($temporalSmartDataContextId);
echo "Updated temporal SmartDataContext: \n" . json_encode($temporalSmartDataContext, JSON_PRETTY_PRINT) . "\n";

# To retrieve an unstructured content from a SmartDataContent
$client->saveUnstructuredDataToFile($temporalSmartDataContextId, $storageObjectId, 'data_retrieved.bin');
if (hash_file('md5', 'data.bin') === hash_file('md5', 'data_retrieved.bin')) {
    echo "Retrieved file data.bin is identical\n";
} else {
    throw new Exception("Retrieved file data.bin is not identical");
}

$client->saveUnstructuredDataToFile($temporalSmartDataContextId, $storageObjectId2, 'data2_retrieved.bin');
if (hash_file('md5', 'data2.bin') === hash_file('md5', 'data2_retrieved.bin')) {
    echo "Retrieved file data2.bin is identical\n";
} else {
    throw new Exception("Retrieved file data2.bin is not identical");
}

unlink('data_retrieved.bin');
unlink('data2_retrieved.bin');
unlink('data.bin');
unlink('data2.bin');

# To remove a unstructured data
$temporalSmartDataContext = $client->removeUnstructuredData($temporalSmartDataContextId, $storageObjectId);
echo "Removed unstructured data with id {$storageObjectId}. Unstructured count: " . count($temporalSmartDataContext['unstructuredData']) . " with id " . $temporalSmartDataContext['unstructuredData'][0]['id'] . "\n";

$temporalSmartDataContext = $client->removeUnstructuredData($temporalSmartDataContextId, $storageObjectId2);
echo "Removed unstructured data with id {$storageObjectId2}. Unstructured count: " . count($temporalSmartDataContext['unstructuredData']) . "\n";
